/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

/**
 * @file tokens.c
 * @brief The implementation for token transmission and reception.
 * This file implements a couple of methods providing functionality
 * to send and receive tokens.
 * @author Salvatore Monforte salvatore.monforte@ct.infn.it
 * @author comments by Marco Pappalardo marco.pappalardo@ct.infn.it and Salvatore Monforte
 */

#include <netdb.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "globus_config.h"
#include "gssapi.h"
#include "gssapi_compat.h"
#include "globus_gss_assist.h"

#include <openssl/x509.h>
#include <openssl/asn1.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>

#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/objects.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>

#define EVP_MD_CTX_init
#define EVP_MD_CTX_cleanup

#define BUFSIZE 4096

#include <time.h>
#include <sys/stat.h>

#include <openssl/buffer.h>
#include <openssl/crypto.h>
#include <openssl/x509v3.h>
#include <openssl/lhash.h>
#include <openssl/rand.h>

#ifdef USE_PKCS11
#include "scutils.h"
#endif

#include <ctype.h>

#include "voms_apic.h"
#include <dirent.h>

#include <signal.h>
#include <sys/wait.h>

struct collection {
  char *user;
  char *userca;
  char *server;
  char *serverca;
  char *voname;
  char *date1;
  char *date2;
  int datalen;
  char *data;
  char *signature;
  int   signed_data;
  int siglen;
  char *uri;
  int type;
  char *custom;
  struct data **std;
  char *copy;
  int signed_len;
};

#include "apic_util.h"


int retrieve(X509 *cert, STACK_OF(X509) *chain, int how, char **buffer, int *buflen, char **vo, int *volen, char **file, int *filelen, int *error)
{
  int size = 0, res;
  int len;
  int found = 0;
  int chain_length;
  X509_EXTENSION *ext;
  int index = 0;
  int nid = 0, nidf = 0, nidv = 0;
  ASN1_OCTET_STRING *obj;

  int j = 0;
  int i;
  int k;
  X509_EXTENSION *ex;
  char buf[1000];
  int ffound = 0;

  if (!cert || !vo || !file || !error || !buffer ||
      (!chain && (how != RECURSE_NONE))) {
    *error = VERR_PARAM;
    return 0;
  }

  /*
   * check credential and get the globus name
   */

  OBJ_create("1.3.6.1.4.1.8005.100.100.1","VOMS","Voms");
  OBJ_create("1.3.6.1.4.1.8005.100.100.2","INCFILE","IncFile");
  OBJ_create("1.3.6.1.4.1.8005.100.100.3","VO","Vo");

  nid  = OBJ_txt2nid("Voms");
  nidf = OBJ_txt2nid("IncFile");
  nidv = OBJ_txt2nid("Vo");

  index = X509_get_ext_by_NID(cert, nidf, -1);
  if (index >= 0) {
    ext = X509_get_ext(cert,index);
    if (ext){
      if (*file = malloc(ext->value->length)) {
	memcpy(*file, ext->value->data,ext->value->length);
	*filelen = ext->value->length;
	found = 1;
      }
    }
  }

  index = X509_get_ext_by_NID(cert, nidv, -1);
  if (index >= 0) {
    ext = X509_get_ext(cert,index);
    if (ext) {
      if (*vo = malloc(ext->value->length)) {
	memcpy(*vo, ext->value->data,ext->value->length);
	*volen=ext->value->length;
	found = 1;
      }
    }
  }

  index = X509_get_ext_by_NID(cert, nid, -1);
  if (index >= 0) {
    ext = X509_get_ext(cert,index);
    if (ext) {
      if (*buffer = malloc(ext->value->length)) {
	memcpy(*buffer, ext->value->data,ext->value->length);
	*buflen = ext->value->length;
	found = 1;
      }
    }
  }

  if (found)
    return 1;

  if (how == RECURSE_CHAIN) {
    /*
     * May need to travel up the chain.
     */

    int position = 0;
      
    chain_length = sk_X509_num(chain);
    while (position < chain_length) {
      cert = sk_X509_value(chain,position);

      index = X509_get_ext_by_NID(cert, nidf, -1);
      if (index >= 0) {
	ext = X509_get_ext(cert,index);
	if (ext){
	  if (*file = malloc(ext->value->length)) {
	    memcpy(*file, ext->value->data,ext->value->length);
	    *filelen = ext->value->length;
	    found = 1;
	  }
	}
      }

      index = X509_get_ext_by_NID(cert, nidv, -1);
      if (index >= 0) {
	ext = X509_get_ext(cert,index);
	if (ext) {
	  if (*vo = malloc(ext->value->length)) {
	    memcpy(*vo, ext->value->data,ext->value->length);
	    *volen=ext->value->length;
	    found = 1;
	  }
	}
      }

      index = X509_get_ext_by_NID(cert, nid, -1);
      if (index >= 0) {
	ext = X509_get_ext(cert,index);
	if (ext) {
	  if (*buffer = malloc(ext->value->length)) {
	    memcpy(*buffer, ext->value->data,ext->value->length);
	    *buflen = ext->value->length;
	    found = 1;
	  }
	}
      }
      if (found)
	return 1;

      position++;
    } 
  }

  *error = VERR_NOEXT;
  return 0;
}




int verify(char *message, int len, struct vomsdata *voms, int *error)
{
  int result = 0;
  char *tmp=message;
  char *subject, *ca;
  struct collection *coll = NULL;
  struct voms *v = NULL;
  struct voms **array;
  ASN1_TIME *time1 = NULL, *time2 = NULL;
  unsigned char *timebuf;

  if (!error)
    return 0;

  if (!message || len <= 0 || !voms || !voms->vdir || !voms->cdir) {
    *error = VERR_NOINIT;
    return 0;
  }

  do {    
    *error = VERR_PARSE;
    coll = unformat(&tmp, len, error);

    if (!coll)
      return 0;

    if (!parse_rest(coll->data, coll->datalen, coll, error))
      goto err;

    *error = VERR_VERIFY;
    if (!check(coll, error, voms->vdir, voms->cdir))
      goto err;

#if 0
    /* Parse the date. */


    timebuf = coll->date1;
    time1 = d2i_ASN1_TIME(NULL, &timebuf, strlen(coll->date1));

    timebuf = coll->date2;
    time2 = d2i_ASN1_TIME(NULL, &timebuf, strlen(coll->date2));

    /* Default error for following checks */
    *error = VERR_TIME;
    if (!time1 || !time2)
      goto err;

    /* Verifies the date constraints */
    if ((X509_cmp_current_time(time1) >=  0) || (X509_cmp_current_time(time2) <= 0))
      goto err;
#endif

    /* Verifies that the cert identifies self */
    if (!id_data(&subject, &ca, error))
      goto err;

    *error = VERR_IDCHECK;
    if (strcmp(coll->user, subject) || strcmp(coll->userca, ca))
      goto err;

    if (v = malloc(sizeof(struct voms))) {
      v->siglen    = coll->siglen;
      v->signature = coll->signature;
      v->user      = coll->user;
      v->userca    = coll->userca;
      v->server    = coll->server;
      v->serverca  = coll->serverca;
      v->voname    = coll->voname;
      v->uri       = coll->uri;
      v->date1     = coll->date1;
      v->date2     = coll->date2;
      v->type      = coll->type;
      v->custom    = coll->custom;
      v->std       = coll->std;
      v->reserved  = message;
      v->datalen   = coll->datalen;

      array=(struct voms **)listadd((char **)(voms->data), (char *)v, sizeof(struct voms *));
      if (array) {
	voms->data = array;
	free(coll->copy);
	free(coll);
	coll = NULL;
      }
      else
	goto err;
    }
    else
      goto err;
  } while (tmp < (message + len));

  result = 1;
 err:
  if (coll) {
    free(coll->copy);
    free(coll);
  }
  if (v && !result)
    free(v);
  if (time1)
    ASN1_TIME_free(time1);
  if (time2)
    ASN1_TIME_free(time2);
  return result;
}

static int
check(struct collection *voms, int *error,  char *vdir, char *cdir)
{
  DIR *dp;
  struct dirent *de;
  int found  = 0;
  int result = 0;
  char *temp  = NULL;

  BIO *in = NULL;
  X509 *x = NULL;

  *error = VERR_DIR;
  if (!(dp = opendir(vdir)))
    return 0;
 
  while (de = readdir(dp)) {
    char *name = de->d_name;

    if (name) {
      in = BIO_new(BIO_s_file());
      if (in) {
	if (temp)
	  free(temp);
	if (temp = malloc(strlen(vdir)+strlen(name)+2)) {
	  strcpy(temp,vdir);
	  strcat(temp,"/");
	  strcat(temp,name);
	}
	if (BIO_read_filename(in, temp) > 0) {
	  x = PEM_read_bio_X509(in, NULL, 0, NULL);
	  if (x) {
	    EVP_PKEY *key = X509_extract_key(x);
	    key = x->cert_info->key->pkey;
	    if (verifysig(key, voms->copy + voms->signed_data, voms->signature,
			  voms->signed_len, voms->siglen)) {
	      found = 1;
	      break;
	    }
	    else X509_free(x);
	  }
	}
      }
    }
  }

  *error = VERR_SIGN;
  if (in) BIO_free(in);
  (void)closedir(dp);

  if (temp)
    free(temp);

  if (found) {
    char buf[2048];

    *error = VERR_SERVER;
    /* check server subject */
    if (strcmp(X509_NAME_oneline(X509_get_subject_name(x), buf,2048),
	       voms->server) == 0) {
      /* check server ca */
      if (strcmp(X509_NAME_oneline(X509_get_issuer_name(x), buf,2048),
		 voms->serverca) == 0) {
	/* check cert */
	result = check_cert(x, error, cdir);
      }
    }
    X509_free(x);
  }

  return result;
}

static int
check_cert(X509 *cert, int *error,  char *cdir)
{
  X509_STORE *ctx = NULL;
  X509_STORE_CTX *csc = NULL;
  X509_LOOKUP *lookup = NULL;
  int i = 0;

  csc = X509_STORE_CTX_new();
  ctx = X509_STORE_new();
  *error = VERR_MEM;
  if (ctx && csc) {
    X509_STORE_set_verify_cb_func(ctx,cb);
    ERR_load_crypto_strings();
#ifdef SIGPIPE
    signal(SIGPIPE,SIG_IGN);
#endif
    CRYPTO_malloc_init();
    SSLeay_add_all_algorithms();
    if (lookup = X509_STORE_add_lookup(ctx, X509_LOOKUP_file())) {
      X509_LOOKUP_load_file(lookup, NULL, X509_FILETYPE_DEFAULT);
      if (lookup=X509_STORE_add_lookup(ctx,X509_LOOKUP_hash_dir())) {
	X509_LOOKUP_add_dir(lookup, cdir, X509_FILETYPE_PEM);
	ERR_clear_error();
	*error = VERR_VERIFY;
	X509_STORE_CTX_init(csc,ctx,cert,NULL);
	i = X509_verify_cert(csc);
      }
    }
  }
  if (ctx) X509_STORE_free(ctx);
  if (csc) X509_STORE_CTX_free(csc);

  if (i)
    return 1;
  else
    return 0;
}


static int MS_CALLBACK 
cb(int ok, X509_STORE_CTX *ctx)
{
  char buf[256];

  if (!ok) {
    X509_NAME_oneline(X509_get_subject_name(ctx->current_cert),buf,256);
    if (ctx->error == X509_V_ERR_CERT_HAS_EXPIRED) ok=1;
    /* since we are just checking the certificates, it is
     * ok if they are self signed. But we should still warn
     * the user.
     */
    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
    /* Continue after extension errors too */
    if (ctx->error == X509_V_ERR_INVALID_CA) ok=1;
    if (ctx->error == X509_V_ERR_PATH_LENGTH_EXCEEDED) ok=1;
    if (ctx->error == X509_V_ERR_INVALID_PURPOSE) ok=1;
    if (ctx->error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT) ok=1;
  }
  return(ok);
}


static int
id_data(char **subj, char **ca, int *error)
{
  OM_uint32                   major_status = 0;
  OM_uint32                   minor_status = 0;

  gss_cred_id_t credential = GSS_C_NO_CREDENTIAL;

  /* acquire our credentials */
  major_status = globus_gss_assist_acquire_cred(&minor_status,
						GSS_C_BOTH,
						&credential);

  if(major_status != GSS_S_COMPLETE) {
    *error = VERR_NOIDENT;
    return 0;
  }

  *ca = getMCA(credential);
  *subj = get_globusid(credential);

  return 1;
}

static char *
get_globusid(gss_cred_id_t handle)
{
  char *globusid;
  char *globusid_tmp;
  gss_name_t server_name = GSS_C_NO_NAME;
  gss_buffer_desc server_buffer_desc = GSS_C_EMPTY_BUFFER;
  gss_buffer_t server_buffer = &server_buffer_desc;
  OM_uint32 major_status = 0;
  OM_uint32 minor_status = 0;
  OM_uint32 minor_status2 = 0;

  if ((major_status = gss_inquire_cred(&minor_status,
				       handle,
				       &server_name,
				       NULL, NULL, NULL)) == GSS_S_COMPLETE) {
    major_status = gss_display_name(&minor_status,
				    server_name, server_buffer, NULL);
    gss_release_name(&minor_status2, &server_name);
  }
  /*
   * The gssapi_cleartext does not implement gss_inquire_cred,
   * so fall back to using environment variable.
   */
  if (major_status == GSS_S_COMPLETE) {
    globusid = (char *)server_buffer_desc.value;
  }
  else {
    return NULL;
  }
  globusid_tmp = strdup(globusid);

  if (server_buffer_desc.value) {
    gss_release_buffer(&minor_status2, server_buffer);
  }
  return globusid_tmp;
}


static int   get_number(char **, char *, char *);
static char *get_name(char **, char *, char *);

/*
 * Function:
 *   get_number(str, name, limit)
 *
 * Description:
 *   This function extracts from the buffer pointed by *'str' a non-zero,
 *   newline-terminated numeric parameter that has the 'name' string has
 *   a prefix.
 *
 * Parameters:
 *   'str'   - A pointer to a pointer to the buffer.  This is done because
 *             the pointer itself will be adjusted in case of successful
 *             execution.
 *   'name'  - The prefix of the number.
 *   'limit' - A pointer to the limit of the buffer.
 *
 * Result:
 *   Failure:
 *     0
 * Success:
 *   The numeric value found. Furthermore, *'str' will be adjusted to point
 *   immediately after the data examined by the function.
 */
static int 
get_number(char **str, char *name, char *limit)
{
  int n = 0;
  char *s;

  if (!str || (!(*str)) || !name || !limit || *str > limit)
    return 0;

  s = *str;

  if (memcmp(s, name, strlen(name)) != 0)
    return 0;

  s += strlen(name);
  if (s > limit)
    return 0;

  while(isdigit(*s)) {
    n *= 10;
    n += *s - '0';
    s++;
    if (s > limit)
      return 0;
  }

  if (*s != '\n' || ((s+1) > limit))
    return 0;

  *str = ++s;
  return n;
}

/*
 * Function:
 *   get_name(str, name, limit)
 *
 * Description:
 *   Totally analogous to get_number, with the only difference that what is
 *   searched for is a newline-terminated string, and NULL is returned in
 *   case of failure.
 *
 * Parameters:
 *   See get_number().
 *
 * Result:
 *   See get_number().
 *
 * NOTE:
 *   This function modifies the buffer. Furthermore, the pointers it
 *   returns are do point INSIDE the modified buffer.
 */
static char *
get_name(char **str, char *name, char *limit)
{
  char *s = *str;
  char *start;

  if (!str || !(*str) || !name || !limit || *str > limit )
    return NULL;

  s = *str;

  if (memcmp(s, name, strlen(name)) != 0)
    return NULL;

  start = s += strlen(name);

  while ((s<=limit) && (*s != '\n'))
    s++;

  if (s > limit)
    return NULL;

  *s = '\0';
  *str = ++s;

  return start;
}

struct collection *
unformat(char **data, int len, int *error)
{
  char *tmp;
  char *limit;
  char *signature = NULL;
  int   signed_data_start;
  char *sig_end;

  int sig_len = 0;
  struct collection *results = NULL;
  char *copy = NULL;

  *error = VERR_PARAM;
  if (!data || (len <= 0))
    return NULL;

  tmp = *data;
  limit = tmp + len - 1;
  *error = VERR_MEM;
  if ((copy = (char *)malloc(len)) == NULL)
    return NULL;
  else
    memcpy(copy, *data, len);

  /*
   * Load the signature.
   */
  sig_len = get_number(&tmp, "SIGLEN:", limit);
  if ((sig_len == 0) || (tmp > limit))
    goto err;

  if (memcmp(tmp, "SIGNATURE:", 10) != 0)
    goto err;

  tmp += 10;
  if (tmp > limit)
    goto err;

  signature = tmp;
  tmp += sig_len;
  signed_data_start = tmp - *data;
  sig_end = tmp;

  if (tmp > limit)
    goto err;

  /*
   * Extracts data from the buffer.
   */

  if ((results = (struct collection *)malloc(sizeof(struct collection))) == NULL)
    return NULL;

  results->user     = get_name  (&tmp, "USER:"   , limit);
  results->userca   = get_name  (&tmp, "UCA:"    , limit);
  results->server   = get_name  (&tmp, "SERVER:" , limit);
  results->serverca = get_name  (&tmp, "SCA:"    , limit);
  results->voname   = get_name  (&tmp, "VO:"     , limit);
  results->uri      = get_name  (&tmp, "URI:"    , limit);
  results->date1    = get_name  (&tmp, "TIME1:"  , limit);
  results->date2    = get_name  (&tmp, "TIME2:"  , limit);
  results->datalen  = get_number(&tmp, "DATALEN:", limit);

  if (!results->user || !results->server || !results->date1 ||
      !results->date2 || !results->uri || results->datalen == 0 ||
      (tmp+results->datalen-1 > limit))
	goto err;

  results->siglen = sig_len;
  results->signed_len  = tmp - *data - signed_data_start + results->datalen;
  results->signature   = signature;
  results->signed_data = signed_data_start;
  results->data = tmp;
  results->copy = copy;

  *data =  results->data + results->datalen;
  return results;

 err:
  if (results)
    free(results);
  if (copy)
    free(copy);
  return NULL;
}

int
parse_rest(char *data, int len, struct collection *voms, int *error)
{
  char *tmp = data;
  char *limit = tmp + len - 1;
  char *group, *role, *cap;
  struct data *d;
  int mistake = 0;

  *error = VERR_PARAM;
  if (!data || (len <= 0) || !voms || !error)
    return 0;

  if (memcmp(tmp, "NO DATA", 7) == 0) {
    voms->type = TYPE_NODATA;
    return 1;
  }
  else if (memcmp(tmp, "GROUP: ", 7)  == 0) {
    *error = VERR_FORMAT;
    voms->type = TYPE_STD;
    do {
      group = get_name(&tmp, "GROUP: ", limit);
      role  = get_name(&tmp, "ROLE:"  , limit);
      cap   = get_name(&tmp, "CAP: "  , limit);

      if (d = (struct data*)malloc(sizeof(struct data))) {
	struct data **dt;
	d->group = group;
	d->role  = role;
	d->cap   = cap;
	dt = (struct data **)listadd(((char **)voms->std),(char *)d,sizeof(struct data *));
	if (!dt) {
	  free(d);
	  mistake = 1;
	}
	else 
	  voms->std = dt;
      }
      else mistake = 1;
    } while (cap);
    return (mistake==0);
  }
  else {
    voms->type = TYPE_CUSTOM;
    voms->custom = data;
  }

  return 1;
}


int 
net_connect(char *hostname, int port, char *contact, char *command, int commandlen, char **buffer, int *buflen)
{
  struct hostent     *he = NULL;
  struct sockaddr_in  remote;
  int                 sock;

  OM_uint32    req_flags, ret_flags;
  OM_uint32    major_status = 0;
  OM_uint32    minor_status = 0;
  OM_uint32    minor_status2 = 0;
  gss_cred_id_t  credential = GSS_C_NO_CREDENTIAL;
  gss_ctx_id_t   context = GSS_C_NO_CONTEXT;
  int          token_status;
  char        *message = NULL;
  unsigned int length;
  int          result = 0;   /* Result of the operation. Failure by default */
  unsigned char value;
  socklen_t len = sizeof(value);
  socklen_t addrlen = sizeof(struct sockaddr_in);
  struct sockaddr_in myaddr_in;

  /* Find and contact the host */
  he = gethostbyname(hostname);
  if (!he) {
    fprintf(stderr, "Host %s not found!\n", hostname);
    return 0;
  }

  remote.sin_addr.s_addr = *(unsigned long*)(he->h_addr_list[0]);
  remote.sin_port=htons(port);
  remote.sin_family = AF_INET;

  sock = socket(AF_INET, SOCK_STREAM, 0);

  if (sock == -1) {
    fprintf(stderr, "Unable to create socket!\n");
    return 0;
  }


  value = 1; /* force reuse  */
  setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *) &value, len );

  if (connect(sock, (struct sockaddr *)&remote, sizeof(remote)) < 0) {
    fprintf(stderr, "Unable to connect to server: %s, port: %d\n", hostname, port);
    close(sock);
    return 0;
  }
  memset ((char *)&myaddr_in, 0, sizeof(struct sockaddr_in));
      
  if (getsockname(sock, (struct sockaddr*)&myaddr_in, &addrlen) == -1)
    return 0;

  /* Initialize GSS */
  major_status == globus_gss_assist_acquire_cred(&minor_status,
						 GSS_C_BOTH,
						 &credential);

  if (major_status != GSS_S_COMPLETE) {
    globus_gss_assist_display_status(stderr,
				     "GSS failed getting user credentials: ",
				     major_status, minor_status, 0);
    goto cleanup;
  }

  /*  req_flags = GSS_C_MUTUAL_FLAG | GSS_C_CONF_FLAG | GSS_C_INTEG_FLAG;*/
  req_flags =0;

  major_status = globus_gss_assist_init_sec_context(&minor_status,
						    credential, &context,
						    contact,
						    req_flags, &ret_flags,
						    &token_status,
						    get_token, &sock,
						    send_token, &sock);

  if (major_status != GSS_S_COMPLETE) {
    globus_gss_assist_display_status(stderr,
				     "GSS failed initializing context: ",
				     major_status, minor_status, token_status);
    goto cleanup;
  }

  if (!((ret_flags & GSS_C_MUTUAL_FLAG) &&
	(ret_flags & GSS_C_CONF_FLAG) &&
	(ret_flags & GSS_C_INTEG_FLAG))) {
    fprintf(stderr, "Confidentiality error with server %s:%d\n",hostname,port);
    goto cleanup;
  }

  /* Send Request. */
  major_status = globus_gss_assist_wrap_send(&minor_status, context,
					     command, commandlen,
					     &token_status,
					     send_token, &sock, stderr);

  if (major_status != GSS_S_COMPLETE) {
    globus_gss_assist_display_status(stderr,
				     "GSS failed initializing context: ",
				     major_status, minor_status, token_status);
    goto cleanup;
  }

  /* Get results. */
  major_status = globus_gss_assist_get_unwrap(&minor_status, context,
					      &message, &length, &token_status,
					      get_token, &sock,  stderr);
  
  if (major_status != GSS_S_COMPLETE) {
    globus_gss_assist_display_status(stderr,
				     "GSS failed receiving data: ",
				     major_status, minor_status, token_status);
    goto cleanup;
  }

  *buffer = malloc(length);
  if (*buffer) {
    memcpy(*buffer, message, length);
    *buflen = length;
    result = 1;
  }

 cleanup:

  if (context != GSS_C_NO_CONTEXT)
    gss_delete_sec_context(&minor_status, &context, GSS_C_NO_BUFFER); 
  if (credential != GSS_C_NO_CREDENTIAL)
    gss_release_cred(&major_status, &credential);
  if (sock != -1)
    close(sock);

  return result;
}   

char **listadd(char **vect, char *data, int size)
{
  int i = 0;
  char **newvect;

  if (!data || (size <= 0))
    return NULL;

  if (vect)
    while (vect[i++]) ;
  else
    i=1;

  if (newvect = (char **)malloc((i+1)*size)) {
    if (vect) {
      newvect[0] = data;
      memcpy(&(newvect[1]), vect, (size*i));
      free(vect);
    }
    else {
      newvect[0] = data;
      newvect[1] = NULL;
    }
    return newvect;
  }
  return NULL;
}

/*
 * The following two functions are an ugly hack of the libglobus_gss*
 * libraries done to let a program discover the CA that issued his own
 * certificate. To do so, you must call the following function. See voms.c
 * for mode of calling.
 */
char *
getMCA(void *credential)
{
  X509 *cert = NULL;
  X509_NAME *name;
  char *result;
  static char *buffer = NULL;
  static int   bufsize = 0;
  int ressize;
  int pos;
  int i;
  int res;
  STACK_OF(X509) *stk;

  if (!credential)
    return NULL;

  pos = sk_X509_num(((gss2_cred_id_desc *)credential)->cred_handle->cert_chain);
  if (pos)
    cert = sk_X509_value(((gss2_cred_id_desc *)credential)->cred_handle->cert_chain,pos-1);
  if (!cert)
    cert = ((gss2_cred_id_desc *)credential)->cred_handle->cert;

  name = X509_get_issuer_name(cert);
  result = X509_NAME_oneline(name, NULL, 0);
  ressize = strlen(result);

  if (ressize >= bufsize) {
    free(buffer);
    buffer = (char *)malloc(strlen(result)+1);
      bufsize=ressize+1;
  }
  if (buffer)
    strncpy(buffer, result, ressize+1);

  OPENSSL_free(result);
  return buffer;
}


#if 0
IMPLEMENT_STACK_OF(X509)
IMPLEMENT_ASN1_SET_OF(X509)

IMPLEMENT_STACK_OF(X509_NAME)

IMPLEMENT_STACK_OF(X509_ATTRIBUTE)
IMPLEMENT_ASN1_SET_OF(X509_ATTRIBUTE)
#endif


/*
 * Function:
 *   verify(key, data, sig_buf, len, sig_len)
 *
 * Description:
 *   This function verifys the signature on a blob of data.
 *
 * Parameters:
 *   'key'     - The key to be used to verify the signature.
 *   'data'    - A pointer to the data that has been signed.
 *   'sig_buf' - A pointer to the signature.
 *   'len'     - The length of the data that has been signed.
 *   'sig_len' - The length of the signature.
 *
 * Result:
 *   Failure:
 *     0.
 *   Success:
 *     1.
 */
int 
verifysig(EVP_PKEY *key, char *data, char *sig_buf, int len, int sig_len)
{
  EVP_MD_CTX     md_ctx;
  int           err;

  if (!key)
    return 0;

  /* Just load the crypto library error strings,
   * SSL_load_error_strings() loads the crypto AND the SSL ones */
  /*  SSL_load_error_strings(); */
  ERR_load_crypto_strings();
  
  /* Verify the signature */
  EVP_MD_CTX_init  (&md_ctx);
  EVP_VerifyInit   (&md_ctx, EVP_sha1());
  EVP_VerifyUpdate (&md_ctx, data, len);
  err = EVP_VerifyFinal (&md_ctx, sig_buf, sig_len, key);
  EVP_MD_CTX_cleanup(&md_ctx);
  
  if (err == 1) {
    return 1;
  }
  else
    return 0;
}



/**
 * Send a gss token.
 * This method send gss tokens using GSI socket objects.
 * @param arg pointer to the descriptor of the socket.
 * @param token pointer to the token buffer to be sent.
 * @param token_length token buffer length
 * @returns the number of bytes sent, or -1 on failure.
 */
int send_token(void *arg, void *token, size_t token_length)
{
    size_t			num_written = 0;
    ssize_t			n_written;
    int 			fd = *( (int *) arg );
    unsigned char		token_length_buffer[4];

    /* encode the token length in network byte order: 4 byte, big endian */
    token_length_buffer[0] = (unsigned char) ((token_length >> 24) & 0xff);
    token_length_buffer[1] = (unsigned char) ((token_length >> 16) & 0xff);
    token_length_buffer[2] = (unsigned char) ((token_length >>  8) & 0xff);
    token_length_buffer[3] = (unsigned char) ((token_length      ) & 0xff);

    /* send the token length */

    while(num_written < 4)
    {
      n_written = send(fd, token_length_buffer + num_written, 4 - num_written,0);
      
      if(n_written < 0)
	{
	  if(errno == EINTR)
	    continue;
	  else
	    return -1;
	}
      else
	num_written += n_written;
    }
    
    /* send the token */

    num_written = 0;
    while(num_written < token_length)
    {
       n_written = send(fd, ((u_char *)token) + num_written, token_length - num_written,0);
       
       if(n_written < 0)
	 {
	   if(errno == EINTR)
		continue;
	   else
	     return -1;
	 }
       else
	 num_written += n_written;
    }
    
    return 0;
}

/**
 * Receive a gss token.
 * This method receives gss tokens using GSI socket objects.
 * @param arg pointer to the descriptor of the socket.
 * @param token pointer to the token buffer to fill with received token.
 * @param token_length token buffer length
 * @returns the number of bytes recieved, or -1 on failure.
 */
int get_token(void *arg, void **token, size_t *token_length)
{
    size_t			num_read = 0;
    ssize_t			n_read;
    int 			fd = *( (int *) arg );
    unsigned char		token_length_buffer[4];

    /* read the token length */

    while(num_read < 4)
    {
      
      n_read = recv(fd,token_length_buffer + num_read, 4 - num_read,0);
	
      if(n_read < 0)
	{
	    if(errno == EINTR)
		continue;
	    else
		return -1;
	}
        else if (n_read == 0)
                return GLOBUS_GSS_ASSIST_TOKEN_EOF;
	else
	    num_read += n_read;
    }
    num_read = 0;
    /* decode the token length from network byte order: 4 byte, big endian */

    *token_length  = (((size_t) token_length_buffer[0]) << 24) & 0xffff;
    *token_length |= (((size_t) token_length_buffer[1]) << 16) & 0xffff;
    *token_length |= (((size_t) token_length_buffer[2]) <<  8) & 0xffff;
    *token_length |= (((size_t) token_length_buffer[3])      ) & 0xffff;

    if(*token_length > 1<<24)
    {
	/* token too large */
	return -1;
    }

    /* allocate space for the token */

    *((void **)token) = (void *) malloc(*token_length);

    if(*token == NULL)
    {
	return -1;
    }

    /* receive the token */

    num_read = 0;
    while(num_read < *token_length)
    {
      n_read = recv(fd, ((u_char *) (*token)) + num_read,(*token_length) - num_read,0);
	
	if(n_read < 0)
	{
	    if(errno == EINTR)
		continue;
	    else
		return -1;
	}
	else
	    if(n_read == 0) return -1; 
	    num_read += n_read;
    }

    return 0;
}
