#include "../../std.h"

#include "../../misc/log_time.h"
#include "../../misc/inttostring.h"
#include "../../config/environment.h"

#include "../httpsd.h"

static int curr_connections = 0;
static int pre_connections = 2;
static int max_connections = 100;

static char* service_not_found_str = "Service not found\r\n";

list<HTTPS_Connector*> HTTPS_Connector::cons;
pthread_mutex_t HTTPS_Connector::cons_lock = PTHREAD_MUTEX_INITIALIZER;

void HTTPS_Connector::loop(void) {
  char buf[1024];
  int keep_alive = 1;
  for(;;) {
    /* read first line of request header */
    size_t l = readline(buf,sizeof(buf)-1);
    if(l == 0) break;
    // parse request
    char* http_URI;
    char* http_version;
    char* http_method = buf;
    char* t = buf;
    t=strchr(t,' ');
    if(t == NULL) {
      olog<<"Broken HTTP request (no URI found) - "<<buf<<endl; break;
    };
    (*t)=0; t++; for(;(*t)==' ';t++);
    http_URI=t;
    t=strchr(t,' ');
    if(t == NULL) {
      olog<<"Broken HTTP request (no version found) - "<<http_URI<<endl; break;
    };
    (*t)=0; t++; for(;(*t)==' ';t++);
    if(strncmp("HTTP/",t,5) != 0) {
      olog<<"Unknown protocol: "<<t<<endl; break;
    };
    http_version=t+5;
    if(keep_alive) {
      if(strcmp(http_version,"1.0") == 0) {
        keep_alive=1;
      } else {
        keep_alive=2;
      };
    };
    // obtain service to use 
    HTTP_Error err;
    HTTP_Service* service = services.get(*this,http_URI);
    if(service == NULL) {
      olog<<"Service not found: "<<http_URI<<endl;
      err=HTTP_NOT_FOUND;
    } else {
      if(strcasecmp(http_method,"GET") == 0) {
        cerr<<"Calling GET"<<endl;
        err=service->get(keep_alive);
      } else if(strcasecmp(http_method,"PUT") == 0) {
        cerr<<"Calling PUT"<<endl;
        err=service->put(keep_alive);
      } else if(strcasecmp(http_method,"POST") == 0) {
        cerr<<"Calling POST"<<endl;
        err=service->post(keep_alive);
      } else {
        cerr<<"Unknown method: "<<http_method<<endl;
        err=HTTP_NOT_IMPLEMENTED;
      };
    };
    if((err==HTTP_NOT_IMPLEMENTED) || (err==HTTP_NOT_FOUND)) { 
      if(skip_header(*this,keep_alive) != 0) break;
    };
    cerr<<"keep_alive: "<<keep_alive<<endl;
    if(err != HTTP_OK) { /* send error response */
      if(send_error_response(*this,keep_alive,err,NULL,NULL) != 0) {
        cerr<<"Failed sending response"<<endl; break;
      };
    };
    if(keep_alive == 0) break;
  };
  olog<<"Exiting\n";
  // close connection
  services.close(*this);
}

void HTTPS_Connector::accept_cb(void *callback_arg,globus_io_handle_t *handle,globus_result_t result) {
  if(result != GLOBUS_SUCCESS) {
    char* tmp = globus_object_printable_to_string(globus_error_get(result));
    cerr<<"Failed to accept connection: "<<tmp<<endl;
    free(tmp);
    return;
  };
  HTTPS_Connector* it = (HTTPS_Connector*)callback_arg;
  globus_thread_blocking_will_block();  // instead of creating new thread
  it->loop();
  delete it;
}

HTTPS_Connector::HTTPS_Connector(globus_io_handle_t *s,HTTP_Services& serv,list<AuthEvaluator*>& auths_):services(serv),auths(auths_) {
  pthread_mutex_lock(&cons_lock);
  cons.push_back(this);
  pthread_mutex_unlock(&cons_lock);
  initialized=false;
  globus_result_t res;

  res=globus_io_tcp_register_accept(s,GLOBUS_NULL,&h,&accept_cb,this);
  if(res != GLOBUS_SUCCESS) {
    char* tmp = globus_object_printable_to_string(globus_error_get(res));
    cerr<<"Failed to accept connection: "<<tmp<<endl; free(tmp);
    return;
  };
  initialized=true;
  return;
}

HTTPS_Connector::~HTTPS_Connector(void) {
  pthread_mutex_lock(&cons_lock);
  for(list<HTTPS_Connector*>::iterator i = cons.begin();i!=cons.end();++i) {
    if((*i)==this) { cons.erase(i); break; };
  };
  pthread_mutex_unlock(&cons_lock);
  if(initialized) globus_io_close(&h);
}

void HTTPS_Connector::identity(globus_io_handle_t* handle,const char* subject,gss_cred_id_t cred) {
  pthread_mutex_lock(&cons_lock);
  for(list<HTTPS_Connector*>::iterator i = cons.begin();i!=cons.end();++i) {
    if(&((*i)->h) == handle) {
      (*i)->user.set(subject,cred);
      cerr<<"Subject is "<<(*i)->user.DN()<<endl;
      cerr<<"Proxy stored to "<<(*i)->user.proxy()<<endl;
      break;
    };
  };
  pthread_mutex_unlock(&cons_lock);
}

size_t HTTPS_Connector::readline(char* buf, size_t l) {
// cerr<<"readline: "<<l<<endl;
  char c;
  unsigned int ll = 0;
  for(;;) {
    if(read(&c,1) != 1) return 0;
    if(ll<l) buf[ll]=c;
    ll++;
    if(c == 0x0a) break;
  };
  if(ll > l) ll=l;
  l=ll-1;
  for(;l>=0;l--) {
    if((buf[l] != 0x0a) && (buf[l] != 0x0d)) break;
    buf[l]=0;
  };
// cerr<<"readline: "<<buf<<endl;
  return ll;
}

size_t HTTPS_Connector::read(char* buf, size_t l) {
  globus_result_t res;
  globus_size_t ll = 0;
  if((res=globus_io_read(&h,(globus_byte_t*)buf,l,1,&ll)) != GLOBUS_SUCCESS) {
    char* tmp = globus_object_printable_to_string(globus_error_get(res));
//    fprintf(stderr,"frecv: %s\n",tmp); free(tmp);
    return 0;
  };
  for(int ii=0;ii<ll;ii++) fprintf(stdout,"%c",buf[ii]);
  return ll;
}

int HTTPS_Connector::write(const char* buf, size_t l) {
  size_t n = 0;
//  for(int ii=0;ii<l;ii++) fprintf(stdout,"%c",buf[ii]);
  for(;n<l;) {
    globus_size_t ll = 0;
    globus_result_t res;
    if((res=globus_io_write(&h,(globus_byte_t*)buf+n,l-n,&ll)) != GLOBUS_SUCCESS) {
      char* tmp = globus_object_printable_to_string(globus_error_get(res));
      fprintf(stderr,"fsend: %s\n",tmp); free(tmp);
      return -1;
    };
    n+=ll;
  };
  return 0;
}

int skip_header(HTTPS_Connector &c,int &keep_alive) {
/* TODO - control keep_alive */
  char buf[1024], *s;
  for (;;) {
    if(c.readline(buf,1023) == 0) return 1; /* connection broken */
//    buf[1023]=0; cerr<<"skip: "<<buf<<endl;
    if(buf[0] == 0) break; /* empty line */
  };
  return 0;
}

int send_error_response(HTTPS_Connector &c,int keep_alive,int code,char* type,char* content) {
  unsigned long long int size = 0;
  if(content) {
    size=strlen(content);
    if(send_response_header(c,keep_alive,code,type,size) != 0) return -1;
    return c.write(content,strlen(content));
  } else {
    string fname = nordugrid_loc+"/share/error"+inttostring(code)+".html";
    stat_file(fname.c_str(),size);
    if(send_response_header(c,keep_alive,code,"text/html",size) != 0) return -1;
    if(size) return send_file(fname.c_str(),c);
  };
  return 0;
}

int send_response_header(HTTPS_Connector &c,int keep_alive,int code,char* type,int size) {
  string header = "HTTP/1.1 "+inttostring(code)+" OK\r\n";
  if(keep_alive && (size>=0)) {
    header+="Connection: Keep-Alive\r\n";
  } else {
    header+="Connection: close\r\n";
  };
  if(size>=0) header+="Content-Length: "+inttostring(size)+"\r\n";
  if(type) { header+="Content-Type: "; header+=type; header+="\r\n"; };
  header+="\r\n";
  olog<<"Sending response:\n"<<header<<endl;
  return c.write(header.c_str(),header.length());
}


