/*
  Download files specified in job.ID.input and check if user uploaded files.
  result: 0 - ok, 1 - unrecoverable error, 2 - potentially recoverable.
  arguments: job_id control_directory session_directory [cache_directory [cache_data_directory [cache_link_directory]]]
  optional arguments: -t threads , -u user
*/
#include "../std.h"
#include <globus_common.h>
#include <globus_gass_copy.h>
#include <globus_ftp_client.h>
#include <globus_object.h>
#include "../files/info_types.h"
#include "../files/info_files.h"
// #include <zlib.h>
// #include "../external/32bitCRC.h"
#include "../misc/checksum.h"
#include "../misc/delete.h"
#include "../misc/url_options.h"
#include "../misc/stringtoint.h"
#include "../misc/log_time.h"
#include "../config/config_map.h"
#include "check.h"
#include "../cache/cache.h"
#include "cache_download.h"
#include "proxy.h"
#include "convert_url.h"

/* default timeout is 5 minutes per file - now 20 */
#define DEFAULT_TIMEOUT 1200
/* check for user uploaded files every 10 seconds */
#define CHECK_PERIOD 10
/* maximum number of retries (for all files together) */
#define MAX_RETRIES 5
/* maximum number simultaneous downloads */
#define MAX_DOWNLOADS 5
typedef struct {
  globus_gass_copy_handle_t handle;
  globus_ftp_client_operationattr_t ftp_attr;
  globus_gass_copy_attr_t o_attr;
  FileDataRC::iterator di;
  bool active;
  string lurl; /* local */
  string rurl; /* remote */
  string cache_url;
  bool cacheable;
} download_handle;
download_handle handles[MAX_DOWNLOADS];

globus_mutex_t list_lock;
globus_mutex_t download_lock;
globus_cond_t  download_cond;
list<FileDataRC> job_files;
JobDescription desc;
JobUser user;
char* cache_dir;
char* cache_data_dir;
char* cache_link_dir;
uid_t cache_uid;
gid_t cache_gid;
char* id;
int download_done=0;
FileDataRC::iterator next_di;
bool leaving = false;

/* if != 0 - change owner of downloaded files to this user */
uid_t file_owner = 0;
uid_t file_group = 0;

bool is_checksum(string str) {
  /* check if have : */
  int n;
  const char *s = str.c_str();
  if((n=str.find('.')) == string::npos) return false;
  if(str.find('.',n+1) != string::npos) return false;
  for(int i=0;i<n;i++) { if(!isdigit(s[i])) return false; };
  for(int i=n+1;s[i];i++) { if(!isdigit(s[i])) return false; };
  return true;
}

int clean_files(list<FileDataRC> &job_files,char* session_dir) {
  string session(session_dir);
  /* delete only downloadable files, let user manage his/hers files */
  list<FileData> tmp;
  if(delete_all_files(session,(list<FileData>&)job_files,false,true,false) != 2) return 0;
  return 1;
}

/*
   Check for existence of user uploadable file
   returns 0 if file exists 
           1 - it is not proper file or other error
           2 - not here yet 
*/
int user_file_exists(FileData &dt,char* session_dir) {
  struct stat st;
  string fname=string(session_dir) + '/' + dt.pfn;
  /* check if file does exist at all */
  if(lstat(fname.c_str(),&st) != 0) return 2;
  /* check for misconfiguration */
  if(!S_ISREG(st.st_mode)) return 1;
  /* parse files information */
  const char *str = dt.lfn.c_str();
  char *str_;
  unsigned long long int fsize;
  unsigned long long int fsum;
  bool have_size = false;
  bool have_checksum = false;
  fsize = strtouq(str,&str_,10);
  if((*str_) == '.') {
    if(str_ != str) have_size=true;
    str=str_+1;
    fsum = strtouq(str,&str_,10);
    if((*str_) != 0) {
      olog << "Invalid checksum in " << dt.lfn << " for " << dt.pfn << endl;
      return 1;
    };
    have_checksum=true;
  }
  else {
    if(str_ != str) have_size=true;
    if((*str_) != 0) {
      olog << "Invalid file size in " << dt.lfn << " for " << dt.pfn << endl;
      return 1;
    };
  };
  /* now check if proper size */
  if(have_size) {
    if(st.st_size < fsize) return 2;
    if(st.st_size > fsize) {
      olog << "Invalid file " << dt.pfn << " is too big." << endl;
      return 1; /* too big file */
    };
  };
  if(have_checksum) {
    int h=open(fname.c_str(),O_RDONLY);
    if(h==-1) { /* if we can't read that file job won't too */
      olog << "Error accessing file " << dt.pfn << endl;
      return 1;
    };
    CheckSum crc;
    char buffer[1024];
    ssize_t l;
    size_t ll = 0;
    crc.start();
    for(;;) {
      if((l=read(h,buffer,1024)) == -1) {
        olog << "Error reading file " << dt.pfn << endl; return 1;
      };
      if(l==0) break; ll+=l;
      crc.add(buffer,l);
    };
    close(h);
    if(fsum != crc.end()) {
      if(have_size) { /* size was checked - it is an error to have wrong crc */ 
        olog << "File " << dt.pfn << " has wrong CRC." << endl;
        return 1;
      };
      return 2; /* just not uploaded yet */
    };
  };
  return 0; /* all checks passed - fiel is ok */
}

void download_cancel_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {
  download_handle *h = (download_handle*)arg;
  globus_mutex_lock(&(list_lock));
  h->active=false;
  globus_mutex_unlock(&(list_lock));
};

void download_finished_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {
  download_handle *h = (download_handle*)arg;
   olog << "Transfer callback: " << h->di->lfn << " (" << h->di->pfn << ")" << endl;
  globus_mutex_lock(&(list_lock));
  if(error != GLOBUS_SUCCESS) {
    olog << "Transfer failed: " << h->di->lfn << " (" << h->di->pfn << ")" << endl;
    char* tmp = globus_object_printable_to_string(error);
    olog<<"Error is: "<<tmp<<endl; free(tmp);
    h->lurl.resize(0);
    h->active=false;
    if(h->cacheable) {
      cache_release_url(cache_dir,cache_data_dir,cache_uid,cache_gid,
                                  h->cache_url.c_str(),id,true);
    };
    h->cache_url.resize(0);
  }
  else {
    if(!leaving) { /* canceling transfer can appear here as success */
      olog << "Downloaded: " << h->di->lfn << " (" << h->di->pfn << ")" << endl;
      /* set proper owner */
      if(getuid() == 0) {
        uid_t uid = (uid_t)-1;
        gid_t gid = (gid_t)-1;
        if(file_owner != 0) uid=file_owner;
        if(file_group != 0) gid=file_group;
        const char* path = get_url_path(h->lurl.c_str());
        if(path) lchown(path,uid,gid);
      };
      /* make sure erasing does not invalidate current pointer */
      if(next_di == h->di) { next_di=job_files.erase(h->di); }
      else { job_files.erase(h->di); };
      /* save changed list */
      if(!job_input_write_file(desc,user,(list<FileData>&)job_files)) {
        olog << "Failed writing changed input file." << endl;
      };
    }
    else {
      if(h->cacheable) {
        cache_release_url(cache_dir,cache_data_dir,cache_uid,cache_gid,
                                    h->cache_url.c_str(),id,true);
      };
    };
    h->lurl.resize(0);
    h->cache_url.resize(0);
    h->active=false;
  };
  globus_mutex_unlock(&(list_lock));
  globus_mutex_lock(&(download_lock));
  download_done=1;
  globus_cond_signal(&download_cond);
  globus_mutex_unlock(&(download_lock));
}

/* this function will wake up main routine periodically.
   it's main purpose is to check for user uploaded files */
void* timer_func(void* arg) {
  int timeout = (int)arg;
  for(;;) {
    sleep(timeout);
    globus_mutex_lock(&(download_lock));
    download_done=2;
    globus_cond_signal(&download_cond);
    globus_mutex_unlock(&(download_lock));
  };
};

/* convert between types of file storage */
void MinusPlus(list<FileData> &from, list<FileDataRC> &to) {
  for(FileData::iterator i=from.begin();i!=from.end();++i) {
    to.push_back(FileDataRC(*i));
  };
//  from.erase();
}

int main(int argc,char** argv) {
  int res=0;
  bool new_download;
  bool not_uploaded;
  int retries=0;
  globus_thread_t timer_thread;
  time_t start_time=time(NULL);
  time_t download_timeout;
  int n_threads = 1;
  LogTime::active=true;
  cache_uid=0;
  cache_gid=0;
  cache_dir=NULL;
  cache_link_dir=NULL;

  // process optional arguments
  for(;;) {
    opterr=0;
    char optc=getopt(argc,argv,"+ht:u:U:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"downloader [-t threads] job_id control_directory session_directory [cache_directory]"<<endl; 
        exit(1);
      }; break;
      case 't': {
        n_threads=atoi(optarg);
        if(n_threads < 1) {
          olog<<"Wrong number of threads"<<endl; exit(1);
        };
      }; break;
      case 'U': {
        unsigned int tuid;
        if(!stringtoint(string(optarg),tuid)) {
          olog<<"Bad number "<<optarg<<endl; exit(1);
        };
        struct passwd pw_;
        struct passwd *pw;
        char buf[BUFSIZ];
        getpwuid_r(tuid,&pw_,buf,BUFSIZ,&pw);
        if(pw == NULL) {
          olog<<"Wrong user name"<<endl; exit(1);
        };
        file_owner=pw->pw_uid;
        file_group=pw->pw_gid;
        if((getuid() != 0) && (getuid() != file_owner)) {
          olog<<"Specified user can't be handled"<<endl; exit(1);
        };
      }; break;
      case 'u': {
        struct passwd pw_;
        struct passwd *pw;
        char buf[BUFSIZ];
        getpwnam_r(optarg,&pw_,buf,BUFSIZ,&pw);
        if(pw == NULL) {
          olog<<"Wrong user name"<<endl; exit(1);
        };
        file_owner=pw->pw_uid;
        file_group=pw->pw_gid;
        if((getuid() != 0) && (getuid() != file_owner)) {
          olog<<"Specified user can't be handled"<<endl; exit(1);
        };
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };
  // process required arguments
  id = argv[optind+0];
  if(!id) { olog << "Missing job id" << endl; return 1; };
  char* control_dir = argv[optind+1];
  if(!control_dir) { olog << "Missing control directory" << endl; return 1; };
  char* session_dir = argv[optind+2];
  if(!session_dir) { olog << "Missing session directory" << endl; return 1; };
  cache_dir = argv[optind+3];
  if(cache_dir) {
    olog<<"Cache directory is "<<cache_dir<<endl;
    cache_data_dir=argv[optind+4];
    if(!cache_data_dir) { 
      cache_data_dir=cache_dir;
    }
    else {
      olog<<"Cache data directory is "<<cache_data_dir<<endl;
      cache_link_dir=argv[optind+5];
      if(cache_link_dir) olog<<"Cache links will go to directory "<<cache_link_dir<<endl;
    };
  };

  prepare_proxy();

  if(n_threads > 10) {
    olog<<"Won't use more than 10 threads"<<endl;
    n_threads=10;
  };
  UrlMapConfig url_map;
  olog<<"Downloader started"<<endl;
  if(globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
    olog<<"COMMON module activation failed\n"; exit(1);
  };
  if(globus_module_activate(GLOBUS_FTP_CLIENT_MODULE) != GLOBUS_SUCCESS) {
    olog<<"FTP_CLIENT module activation failed\n";
    globus_module_deactivate(GLOBUS_COMMON_MODULE); exit(1);
  };
  if(globus_module_activate(GLOBUS_GASS_COPY_MODULE) != GLOBUS_SUCCESS) {
    olog<<"GASS_COPY module activation failed\n";
    globus_module_deactivate(GLOBUS_FTP_CLIENT_MODULE);
    globus_module_deactivate(GLOBUS_COMMON_MODULE); exit(1);
  };
  // prepare Job and User descriptions
  desc=JobDescription(id,session_dir);
  uid_t uid;
  gid_t gid;
  if(file_owner != 0) { uid=file_owner; }
  else { uid= getuid(); };
  if(file_group != 0) { uid=file_group; }
  else { gid= getgid(); };
  cache_uid=getuid();
  cache_gid=getgid();
  user=JobUser(uid);
  user.SetControlDir(control_dir);
  user.SetSessionRoot(session_dir); /* not needed */
  if(cache_dir) user.SetCacheDir(cache_dir,cache_data_dir);
  // get the list of input files
  list<FileData> job_files_;
  if(!job_input_read_file(desc.get_id(),user,job_files_)) {
    olog << "Can't read list of input files" << endl; res=1; goto exit;
  };
  MinusPlus(job_files_,job_files);
  download_timeout=DEFAULT_TIMEOUT * (job_files.size() + 1);
  // remove bad files
  if(clean_files(job_files,session_dir) != 0) {
    olog << "Can't remove junk files" << endl; res=1; goto exit;
  };
  // initialize structures to handle download
  int i;
  for(i=0;i<MAX_DOWNLOADS;i++) {
    if(globus_gass_copy_handle_init(&(handles[i].handle),GLOBUS_NULL) != GLOBUS_SUCCESS) {
      olog << "Failed to init globus handle" << endl; res=2; goto exit;
    };
    globus_ftp_control_parallelism_t ftp_paral;
//    globus_ftp_control_dcau_t ftp_dcau;
//    ftp_dcau.mode=GLOBUS_FTP_CONTROL_DCAU_DEFAULT;
//    ftp_dcau.subject=;
    globus_ftp_client_operationattr_init(&(handles[i].ftp_attr));
//    globus_ftp_client_operationattr_set_dcau(&(handles[i].ftp_attr,&ftp_dcau);
    if(n_threads>1) {
      olog<<"Using default parallel download !"<<endl;    
      ftp_paral.fixed.mode=GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
      ftp_paral.fixed.size=n_threads;
      globus_ftp_client_operationattr_set_parallelism(
                          &(handles[i].ftp_attr),&ftp_paral);
    };
    globus_ftp_client_operationattr_set_striped(
               &(handles[i].ftp_attr),GLOBUS_FALSE);
    globus_ftp_client_operationattr_set_type(
               &(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_TYPE_IMAGE);
    globus_ftp_client_operationattr_set_mode(
               &(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
    globus_ftp_client_operationattr_set_data_protection(
               &(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    globus_ftp_client_operationattr_set_control_protection(
               &(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    globus_gass_copy_attr_init(&(handles[i].o_attr));
    globus_gass_copy_attr_set_ftp(&(handles[i].o_attr),&(handles[i].ftp_attr));
    handles[i].active=false;
  };
  globus_mutex_init(&list_lock,GLOBUS_NULL);
  globus_mutex_init(&download_lock,GLOBUS_NULL);
  globus_cond_init(&download_cond,GLOBUS_NULL);
  if(globus_thread_create(&timer_thread,NULL,&timer_func,(void*)CHECK_PERIOD) != 0) {
    olog << "Failed to start new thread" << endl; res=2; goto exit;
  };
  download_done=0; 
  next_di=job_files.begin();
  new_download=false;
  not_uploaded=false;
  // run cycle trying to download files
  for(;;) {
    // find something to download
    globus_mutex_lock(&(list_lock)); // lock list
    if(next_di == job_files.end()) { // no more files to try - wrap
      if(new_download) { // last cycle inited download, increase retries counter
        new_download=false;
        /* check retry counter */
        retries++; 
        if(retries >= MAX_RETRIES) { 
          olog << "Maximum number of retries exceeded." << endl;
          globus_mutex_unlock(&(list_lock)); res=2; break;
        };
      };   
      /* no more tries or nothing to download */
      if(!not_uploaded) { // user uploaded everything
        /* check for pending downloads */
        for(i=0;i<MAX_DOWNLOADS;i++) { if(handles[i].active) break; };
        if(i >= MAX_DOWNLOADS) { /* all downloads completed or out of tries */
          olog << "No current downloads found" << endl;
          globus_mutex_unlock(&(list_lock)); break;
        };
      }
      else {
        not_uploaded=false;
      };
      i=MAX_DOWNLOADS; /* mark - just wait */
    }
    else {
      if(next_di->lfn.find(":") == string::npos) { /* is it lfn */
        /* process user uploadable file */
        olog << "Check user uploadable file: " << (*next_di).lfn << endl;
        /* user uploadable file - check it */
        int err=user_file_exists(*next_di,session_dir);
        if(err == 0) { /* file is uploaded */
          olog << "User has uploaded file" << endl;
          next_di=job_files.erase(next_di);
          if(!job_input_write_file(desc,user,(list<FileData>&)job_files)) {
            olog << "Failed writing changed input file." << endl;
          };
        }
        else if(err == 1) { /* critical failure */
          olog << "Critical error for uploadable file" << endl;
          globus_mutex_unlock(&(list_lock)); res=1; break;
        }
        else {
          not_uploaded=true; ++next_di;
        };
        globus_mutex_unlock(&(list_lock)); continue;
      };
      /* isn't it already being downloaded */
      for(i=0;i<MAX_DOWNLOADS;i++) { 
        if((handles[i].active) && (handles[i].di == next_di)) break;
      };
      if(i < MAX_DOWNLOADS) {
        ++next_di; globus_mutex_unlock(&(list_lock)); continue;
      };
      /* have file to download - find unused handle */
      for(i=0;i<MAX_DOWNLOADS;i++) { if(!handles[i].active) break; };
    };
    if(i >= MAX_DOWNLOADS) {  /* no handles to use or no new downloads */
      olog << "Waiting till download finishes ("<<res<<")" << endl;
      globus_mutex_unlock(&(list_lock)); /* release list */
      /* just wait till download finishes */
/*
      struct timespec ts;
      ts.tv_sec=CHECK_PERIOD*1000;
      ts.tv_nsec=1;
*/
      globus_mutex_lock(&(download_lock));
      while(!(download_done)) {
        if(globus_cond_wait(&(download_cond),&(download_lock)) ==
           ETIMEDOUT) break;
      };
      download_done=0;
      globus_mutex_unlock(&(download_lock));
      /* if at the end of list - wrap */
      globus_mutex_lock(&(list_lock));
      if(next_di==job_files.end()) next_di=job_files.begin();
      globus_mutex_unlock(&(list_lock));
      /* check for gloabal timeout */
      if((time(NULL) - start_time) > download_timeout) {
        olog << "Download/upload timeout" << endl; res=2; break;
      };
      continue;
    };
    /* have place and file to download */
    handles[i].di=next_di; handles[i].active=true; /* claim handle */
    /* define place to store */
    handles[i].lurl=string("file://") + session_dir + (*next_di).pfn;
    /* handles[i].rurl=next_di->lfn; */
    /* process RC urls, also sort urls with local access first,
       rurl is filled from next_di->lfn */
    if(!convert_url(handles[i].rurl,*next_di,url_map,true,false)) {
      olog << "URL resolving problem." << endl; res=2;
      globus_mutex_unlock(&(list_lock)); break;
    };
    olog<<"Going to start download: "<<next_di->lfn<<
          " to "<<handles[i].lurl<<endl;
    globus_result_t error;
    globus_gass_copy_attr_t *o_attr_p = NULL;
    /* setting number of download threads */
    string thr_num_s;
    /* define nuber of parallel threads */
    int thr_num = n_threads; /* default value given in options */
    if(get_url_option(handles[i].rurl,"threads",0,thr_num_s) == 0) {
      if(stringtoint(thr_num_s,thr_num)) {
        if(thr_num < 1) thr_num = 1;
        if(thr_num > 10) thr_num = 10;
      }
      else { thr_num = 1; };
    };
    /* check for cachebility */
    handles[i].cacheable=false;
    if(cache_dir != NULL) {
      string cache_s;
      if((get_url_option(handles[i].rurl,"cache",0,cache_s) != 0) ||
         (strcasecmp("no",cache_s.c_str()) != 0)) { /* allowed to cache */
          handles[i].cacheable=true;
      };
    };
    /* strip url from options */
    remove_url_options(handles[i].rurl);
    /* mapping url locally - maybe */
    url_map.map(handles[i].rurl);
    /* link:/ urls can't be cached. This also has no sense. */
    if(strncasecmp("link:/",handles[i].rurl.c_str(),6) == 0) {
      handles[i].cacheable=false;
    };
    globus_ftp_control_parallelism_t ftp_paral;
    if(thr_num <= 1) {
      ftp_paral.mode=GLOBUS_FTP_CONTROL_PARALLELISM_NONE;
      ftp_paral.base.size=1;
    }
    else {
      ftp_paral.mode=GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
      ftp_paral.fixed.size=thr_num;
    };
    globus_ftp_client_operationattr_set_parallelism(&(handles[i].ftp_attr),&ftp_paral);
    if(!strncmp(next_di->lfn.c_str(),"gsiftp://",9)) {
      o_attr_p = &(handles[i].o_attr);
      olog<<"Using "<<thr_num<<" threads"<<endl;
    };
    if(handles[i].cacheable) {
      /* initiate download to  cache */
      handles[i].cache_url=next_di->lfn;
      canonic_url(handles[i].cache_url);
      if(cache_find_url(cache_dir,cache_data_dir,cache_uid,cache_gid,
                                  handles[i].cache_url.c_str(),id) != 0) {
        handles[i].cacheable=false;
      }
      else {
        handles[i].cacheable=true;
        if((error=globus_gass_copy_register_url_to_url_cache(
             &(handles[i].handle),
             (char*)(handles[i].rurl.c_str()),o_attr_p,
             (char*)(handles[i].lurl.c_str()),GLOBUS_NULL,
             handles[i].cache_url.c_str(),
             cache_dir,cache_data_dir,id,0,0,cache_link_dir,
             &download_finished_callback,handles+i)) != GLOBUS_SUCCESS) {
            /* not necessary */
          cache_release_url(cache_dir,cache_data_dir,cache_uid,cache_gid,
                                 handles[i].cache_url.c_str(),id,true);
          handles[i].cacheable=false;
          olog<<"Failed to initiate download to cache "<<next_di->lfn<< endl;
          char* tmp=globus_object_printable_to_string(globus_error_get(error));
          olog<<"Error is: "<<tmp<<endl; free(tmp);
        };
      };
    };
    if(!handles[i].cacheable) {
      if((error=(globus_object_t*)globus_gass_copy_register_url_to_url_check(
               &(handles[i].handle),
               (char*)(handles[i].rurl.c_str()),o_attr_p,
               (char*)(handles[i].lurl.c_str()),GLOBUS_NULL,
               &download_finished_callback,handles+i)) != GLOBUS_SUCCESS) {
      /* bad url probably - ???? leave with error ???? */
        olog << "Bad url or local file: " << next_di->lfn << " (" << next_di->pfn << ")" << endl;
        char* tmp=globus_object_printable_to_string(globus_error_get(error));  
        olog<<"Error is: "<<tmp<<endl; free(tmp);
        handles[i].active=false;
        handles[i].lurl.resize(0);
        ++next_di;
        globus_mutex_unlock(&(list_lock));
        res=1; break; /* leaving program with error (unrecoverable)*/
      };
    };
    olog << "Download started ("<<res<<")" << endl;
    new_download=true;
    ++next_di;
    globus_mutex_unlock(&(list_lock));
  };
  /* still can be downloads running - must cancel here */
  olog << "Leaving downloader ("<<res<<")"<<endl;
  leaving=true;
//  globus_mutex_lock(&(list_lock));
//  for(i=0;i<MAX_DOWNLOADS;i++) {
//    if(handles[i].active) {
//      olog << "Canceling connection"<<endl;
//      globus_gass_copy_cancel(&(handles[i].handle),
//                   download_cancel_callback,handles+i);
//    };
//  };
//  globus_mutex_unlock(&(list_lock));
  /*  wait for cancel */
  sleep(3);
olog << "sleep passed ("<<res<<")"<<endl;
  /* clean unfinished files here */
  clean_files(job_files,session_dir);

  for(i=0;i<MAX_DOWNLOADS;i++) {
    globus_gass_copy_handle_destroy(&(handles[i].handle));
    globus_ftp_client_operationattr_destroy(&(handles[i].ftp_attr));
//    globus_gass_copy_attr_destroy(&(handles[i].o_attr));
  };
  globus_mutex_destroy(&download_lock);
  globus_cond_destroy(&download_cond);
  globus_mutex_destroy(&list_lock);
exit:
  if(res != 0) {
    if(cache_dir) cache_release_url(cache_dir,cache_data_dir,cache_uid,cache_gid,id,true);
  };
  remove_proxy();
olog << "Beeing BAD"<<endl;
  exit(res);
olog << "Deactivate "<<endl;
olog << "Deactivate GASS"<<endl;
  globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
olog << "Deactivate FTP"<<endl;
  globus_module_deactivate(GLOBUS_FTP_CLIENT_MODULE);
olog << "Deactivate COMMON"<<endl;
  globus_module_deactivate(GLOBUS_COMMON_MODULE);
olog << "Exit code "<<res<<endl;
  return res;
}
