/*
  upload files from user interface to CE .
  result: 0 - ok, 1 - unrecoverable error, 2 - potentially recoverable.
*/
#include "../std.h"
#include <globus_common.h>
#include <globus_gass_copy.h>
#include <globus_ftp_client.h>
#include <globus_ftp_client_plugin.h>
#include <globus_gram_client.h>
#include <globus_gram_protocol.h>
#include <globus_object.h>
#include "replica.h"
#include "replica_utils.h"
#include "../files/info_types.h"
#include "../files/info_files.h"
#include "../misc/delete.h"
#include "../misc/url_options.h"
#include "../misc/canonical_dir.h"
#include "../misc/log_time.h"
#include "autodir.h"
#include "ftpsubmit.h"
#include "ui_uploader.h"

/* default timeout is 5 minutes per file */
#define DEFAULT_TIMEOUT 300
/* check for user uploaded files every 10 seconds */
#define CHECK_PERIOD 10
/* maximum number of retries (for all files together) */
#define MAX_RETRIES 5
/* maximum number simultaneous uploads */
#define MAX_UPLOADS 5
typedef struct {
  globus_gass_copy_handle_t handle;
  globus_gass_copy_attr_t handle_attr;
  globus_ftp_client_operationattr_t handle_ftp_attr;
  FileData::iterator di;
  bool active;
  string lurl;     /* local url file:///path */
  string rurl;     /* remote url gsiftp://host:port/path */
} upload_handle;
upload_handle handles[MAX_UPLOADS];

globus_mutex_t list_lock;
globus_mutex_t upload_lock;
globus_cond_t  upload_cond;
list<FileData> job_files;
JobId job_id;
int upload_done=0;
FileData::iterator next_di;
bool leaving = false;

void upload_cancel_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {
  upload_handle *h = (upload_handle*)arg;
  odlog(1) << "Upload canceled: "<<h->di->pfn<<" ("<<h->di->lfn<<")"<<endl;
  globus_mutex_lock(&(list_lock));
  h->active=false;
  globus_mutex_unlock(&(list_lock));
};

void upload_finished_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {
  upload_handle *h = (upload_handle*)arg;
  globus_mutex_lock(&(list_lock));
  if(error != GLOBUS_SUCCESS) {
    odlog(1) << "Transfer failed: "<<h->di->pfn<<" ("<<h->di->lfn<<")"<<endl;
    char* tmp = globus_object_printable_to_string(error);
    odlog(1)<<"Error is: "<<tmp<<endl; free(tmp);
    h->lurl.resize(0);
    h->rurl.resize(0);
    h->active=false;
  }
  else {
    if(!leaving) { /* canceling transfer can appear here as success */
      odlog(1) << "Uploaded: "<<h->di->pfn<<" ("<<h->di->lfn<<")"<<endl;
      /* make sure erasing does not invalidate current pointer */
      if(next_di == h->di) { next_di=job_files.erase(h->di); }
      else { job_files.erase(h->di); };
      /* save changed list */
    };
    h->lurl.resize(0);
    h->rurl.resize(0);
    h->active=false;
  };
  globus_mutex_unlock(&(list_lock));
  globus_mutex_lock(&(upload_lock));
  upload_done=1;
  globus_cond_signal(&upload_cond);
  globus_mutex_unlock(&(upload_lock));
}

/* this function will wake up main routine periodically.
   it's main purpose here is to provide timeout for hanged
   connections */
void* timer_func(void* arg) {
  int timeout = (int)arg;
  for(;;) {
    sleep(timeout);
    globus_mutex_lock(&(upload_lock));
    upload_done=2;
    globus_cond_signal(&upload_cond);
    globus_mutex_unlock(&(upload_lock));
  };
};

bool make_lurl(string &path) {
  if(path[0] == '/') {
    path=string("file://")+path;
  }
  else {
    char buf[BUFSIZ];
    if(getcwd(buf,sizeof(buf)) == NULL) {
      odlog(0) << "Current directory path is too long." << endl;
      return false;
    };
    path=string("file://")+buf+"/"+path;
  };
  return true;
};

#ifdef __UI_LIBRARY__
int ui_uploader(const char* resource,const char* rsl,char** job_id,const char* session_url,rsl_action act,const vector<string> &filenames,int debug_level) {
  char* jobid = NULL;
  if(job_id) jobid=(*job_id);
  LogTime::level=debug_level;
#else
int main(int argc,char** argv) {
  const char* resource = NULL;
  const char* rsl = NULL;
  char* jobid = NULL;
  const char* session_url = NULL;
  rsl_action act = RSL_ACTION_REQUEST;
#endif
  int res=0;
  int i;
  bool new_upload;
  int retries=0;
  globus_thread_t timer_thread;
  time_t start_time=time(NULL);
  time_t upload_timeout;
  LogTime::active=false;

#ifndef __UI_LIBRARY__
  for(;;) {
    opterr=0;
    char optc=getopt(argc,argv,"+hcCi:r:R:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        cerr<<"ui-uploader [-r resource] [-R rsl] [-c] [-C] [-i job_id] session_dir_url|session_root_url [list]"<<endl;
        exit(1);
      }; break;
      case 'c': {
        if(act != RSL_ACTION_REQUEST) {
          cerr<<"Only one non-default action can be specified"<<endl;
          exit(1);
        };
        act=RSL_ACTION_CANCEL;
      }; break;
      case 'C': {
        if(act != RSL_ACTION_REQUEST) {
          cerr<<"Only one non-default action can be specified"<<endl;
          exit(1);
        };
        act=RSL_ACTION_CLEAN;
      }; break;
      case 'i': {
        if(jobid != NULL) {
          cerr<<"Only one id can be given"<<endl;
          exit(1);
        };
        jobid=strdup(optarg);
      }; break;
      case 'r': {
        if(resource != NULL) {
          cerr<<"Only one resource can be given"<<endl;
          exit(1);
        };
        resource=optarg;
      }; break;
      case 'R': {
        if(rsl != NULL) {
          cerr<<"Only one rsl can be specified"<<endl;
          exit(1);
        };
        rsl=optarg;
      }; break;
      case '?': {
        cerr<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        cerr<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        cerr<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };
#endif
  leaving=false;
  char* rsl_act = NULL;
  char* rsl_act_templ="&(executable=/bin/echo)(savestate=yes)(action=%s)(jobid=%s)";
  if((act == RSL_ACTION_CANCEL) || (act == RSL_ACTION_CLEAN)) {
    if(!resource) { odlog(0)<<"Missing resource contact string"<<endl; return 1; };
    if(!jobid) { odlog(0)<<"Missing job identifier"<<endl; return 1; };
    rsl_act=(char*)malloc(strlen(rsl_act_templ)+strlen(jobid)+6+1);
    if(!rsl_act) { odlog(0)<<"Memory allocation error"<<endl; return 1; };
    if(act == RSL_ACTION_CLEAN) {
      sprintf(rsl_act,rsl_act_templ,"clean",jobid);
    }
    else {
      sprintf(rsl_act,rsl_act_templ,"cancel",jobid);
    };
    rsl=rsl_act; 
  }
  else if(act == RSL_ACTION_REQUEST) {
#ifndef __UI_LIBRARY__
    session_url = argv[optind];
#endif
    if(!session_url) {
      odlog(0)<< "Missing session directory url" << endl; return 1;
    };
  }
  else {
    odlog(0)<<"This action is not defined"<<endl;
    return 1;
  };
  if(globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
    odlog(0)<<"COMMON module activation failed\n"; return 1;
  };
  if(resource) {
    if(rsl == NULL) {
      odlog(0)<<"Missing RSL"<<endl;
      return 1;
    };
    if(strncmp(resource,"gsiftp://",9)) {
      if(globus_module_activate(GLOBUS_GRAM_CLIENT_MODULE) != GLOBUS_SUCCESS) {
        odlog(0)<<"GRAM_CLIENT module activation failed\n"; 
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return 1;
      };
      char* job_contact;
      int error=globus_gram_client_job_request((char*)resource,rsl,0,NULL,&job_contact);
      if(error != GLOBUS_SUCCESS) {
        odlog(0)<<"RSL submission failed: "<<globus_gram_protocol_error_string(error)<<endl;
        globus_module_deactivate(GLOBUS_GRAM_CLIENT_MODULE);
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return 1;
      };
      globus_module_deactivate(GLOBUS_GRAM_CLIENT_MODULE);
      if(job_contact == NULL) {
        odlog(0)<<"Globus gatekeeper did not return job contact"<<endl;
        globus_module_deactivate(GLOBUS_COMMON_MODULE); return 1;
      };
      if(strncmp("https://",job_contact,8)) {
        odlog(0)<<"Bad contact returned: "<<job_contact<<endl;
        globus_module_deactivate(GLOBUS_COMMON_MODULE); return 1;
      };
      char *s=strchr(job_contact+8,'/');
      if(s == NULL) {
        odlog(0)<<"Bad contact returned: "<<job_contact<<endl;
        globus_module_deactivate(GLOBUS_COMMON_MODULE); return 1;
      };
      s++; /* start of numbers */
      char* s_ = s+(strlen(s)-1);
      if((s == s_) || ((*s_) != '/')) {
        odlog(0)<<"Bad contact returned: "<<job_contact<<endl;
        globus_module_deactivate(GLOBUS_COMMON_MODULE); return 1;
      };
/*  no more / to . - just remove */
      for(s_=s;s_;) { 
        s_=strchr(s,'/');
        if(s_) memmove(s_,s_+1,strlen(s_));
      }; 
      if(jobid == NULL) jobid=strdup(s);
      odlog(1)<<"Assigned job id: "<<jobid<<endl;
#ifdef __UI_LIBRARY__
      if(job_id) if((*job_id) == NULL) (*job_id)=jobid;
#endif
      globus_free(job_contact);
    }
    else { /* submission through gsiftp */
      if(globus_module_activate(GLOBUS_FTP_CONTROL_MODULE) != GLOBUS_SUCCESS) {
        odlog(0)<<"FTP_CONTROL module activation failed\n"; 
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return 1;
      };
      globus_url_t url_;
      if(globus_url_parse(resource,&url_) != GLOBUS_URL_SUCCESS) {
        odlog(0)<<"Bad resource url"<<endl;
        globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return 1;
      };
      char* host = url_.host;
      if(!host) {
        odlog(0)<<"Missing host in resource url"<<endl;
        globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return 1;
      };
      unsigned short port = url_.port;
      if(port == 0) {
        odlog(0)<<"Missing or wrong port in resource url"<<endl;
        globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return 1;
      };
      char* path = url_.url_path;
      if(!path) {
        odlog(0)<<"Missing path in resource url"<<endl;
        globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
        globus_module_deactivate(GLOBUS_COMMON_MODULE);
        return 1;
      };

      if(act == RSL_ACTION_CANCEL) {
        if(!ftpsubmit(host,port,path,ftpsubmit_cancel_req,&jobid)) {
          odlog(0)<<"Cancel request failed"<<endl;
          globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
          globus_module_deactivate(GLOBUS_COMMON_MODULE);
          return 1;
        };
      }
      else if(act == RSL_ACTION_CLEAN) {
        if(!ftpsubmit(host,port,path,ftpsubmit_clean_req,&jobid)) {
          odlog(0)<<"Clean request failed"<<endl;
          globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
          globus_module_deactivate(GLOBUS_COMMON_MODULE);
          return 1;
        };
      }
      else {
        if(!ftpsubmit(host,port,path,rsl,&jobid)) {
          odlog(0)<<"RSL submission failed"<<endl;
          globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
          globus_module_deactivate(GLOBUS_COMMON_MODULE);
          return 1;
        };
      };
      odlog(1)<<"Assigned job id: "<<jobid<<endl;
#ifdef __UI_LIBRARY__
      if(job_id) if((*job_id) == NULL) (*job_id)=jobid;
#endif
      globus_module_deactivate(GLOBUS_FTP_CONTROL_MODULE);
    };
  };
  if(act != RSL_ACTION_REQUEST) {
    if(rsl_act) free(rsl_act);
    globus_module_deactivate(GLOBUS_COMMON_MODULE); return 0;
  };
  /* 2 arguments required: url of session dir and file containing
     list of files to upload (pairs: remote file name + local file name;
     if local file name not given it is the same as remote; if
     list is not given - read from stdin */
#ifndef __UI_LIBRARY__
  char* file_list = argv[optind+1];
#endif
  string base_url=session_url;
  if(base_url.length() == 0) {
    odlog(0)<< "Session directory url is empty" << endl; return 1;
  };
  if(base_url[base_url.length()-1] == '/') {
    base_url.erase(base_url.length()-1,1);
  };
  if(jobid != NULL) base_url=base_url+"/"+jobid;
  if(globus_module_activate(GLOBUS_GASS_COPY_MODULE) != GLOBUS_SUCCESS) {
    odlog(0)<<"GASS COPY module activation failed\n";
    globus_module_deactivate(GLOBUS_COMMON_MODULE); return 1;
  };
#ifndef __UI_LIBRARY__
  if(file_list) {
    if(!job_Xput_read_file(string(file_list),job_files)) {
      odlog(0)<< "Can't read list of files from " << file_list << endl;
      res=1; goto exit;
    };
  }
  else {
    if(!job_Xput_read_file(job_files)) {
      odlog(0)<< "Can't parse list of files" << endl;
      res=1; goto exit;
    };
  };
#else
  for(vector<string>::const_iterator iv=filenames.begin();iv!=filenames.end();++iv) {
    string f1 = *iv;
    string f2; 
    canonical_dir(f1);
    ++iv; if(iv==filenames.end()) {
      job_files.push_back(FileData(f1.c_str(),NULL));
      break;
    };
    f2 = *iv;
    job_files.push_back(FileData(f1.c_str(),f2.c_str()));
  };
#endif
  
  upload_timeout=DEFAULT_TIMEOUT * (job_files.size() + 1);
  for(i=0;i<MAX_UPLOADS;i++) {
    if(globus_gass_copy_handle_init(&(handles[i].handle),GLOBUS_NULL) != GLOBUS_SUCCESS) {
      odlog(0)<< "Failed to init globus handle" << endl; res=2; goto exit;
    };
    globus_gass_copy_attr_init(&(handles[i].handle_attr));
    globus_ftp_client_operationattr_init(&(handles[i].handle_ftp_attr));
    globus_ftp_client_operationattr_set_data_protection(
                                 &(handles[i].handle_ftp_attr),
                                 GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    globus_gass_copy_attr_set_ftp(&(handles[i].handle_attr),
                                  &(handles[i].handle_ftp_attr));
    handles[i].active=false;
  };
  globus_mutex_init(&list_lock,GLOBUS_NULL);
  globus_mutex_init(&upload_lock,GLOBUS_NULL);
  globus_cond_init(&upload_cond,GLOBUS_NULL);
 /* stupid nexus/globus conditions do not have wait with timeout */
  if(globus_thread_create(&timer_thread,NULL,&timer_func,(void*)CHECK_PERIOD) != 0) {
    odlog(0)<< "Failed to start new thread" << endl; res=2; goto exit;
  };
  upload_done=0; 
  next_di=job_files.begin();
  new_upload=false;
  for(;;) {
    /* find something to upload */
    globus_mutex_lock(&(list_lock));
    if(next_di == job_files.end()) {
      if(new_upload) { /* increase retries only if there were new tries */
        new_upload=false;
        /* check retry counter */
        retries++; 
        if(retries >= MAX_RETRIES) { 
          odlog(0)<< "Maximum number of retries exceeded." << endl;
          globus_mutex_unlock(&(list_lock)); res=2; break;
        };
      };   
      /* nothing to upload */
      /* check for pending uploads */
      for(i=0;i<MAX_UPLOADS;i++) { if(handles[i].active) break; };
      if(i >= MAX_UPLOADS) { /* all uploads completed or out of tries */
        globus_mutex_unlock(&(list_lock)); break;
      };
      i=MAX_UPLOADS; /* mark - just wait */
    }
    else {
      /* isn't it already being uploaded */
      for(i=0;i<MAX_UPLOADS;i++) { 
        if((handles[i].active) && (handles[i].di == next_di)) break;
      };
      if(i < MAX_UPLOADS) {
        ++next_di; globus_mutex_unlock(&(list_lock)); continue;
      };
      /* find unused handle */
      for(i=0;i<MAX_UPLOADS;i++) { if(!handles[i].active) break; };
    };
    if(i >= MAX_UPLOADS) {  /* no handles to use or no new uploads */
      odlog(1)<< "Waiting till any upload finishes" << endl;
      /* wait till upload finishes */
      globus_mutex_unlock(&(list_lock));
      globus_mutex_lock(&(upload_lock));
      while(!(upload_done)) {
        globus_cond_wait(&(upload_cond),&(upload_lock));
      };
      upload_done=0;
      globus_mutex_unlock(&(upload_lock));
      globus_mutex_lock(&(list_lock));
      if(next_di==job_files.end())next_di=job_files.begin();
      globus_mutex_unlock(&(list_lock));
      if((time(NULL) - start_time) > upload_timeout) {
        odlog(0)<< "Uploads timeout" << endl; res=2; break;
      };
      continue;
    };
    handles[i].di=next_di; handles[i].active=true;
    if((*next_di).lfn.length() != 0) { handles[i].lurl=(*next_di).lfn; }
    /* pfn always starts with '/' */
    else { handles[i].lurl=(*next_di).pfn.substr(1); };
    if(!make_lurl(handles[i].lurl)) {
      odlog(1)<< "Failed converting local name to local url" << endl;
    };
    handles[i].rurl=base_url+(*next_di).pfn;
odlog(1)<< "Starting upload: " << (*next_di).pfn << " ("<<handles[i].rurl <<") from " << handles[i].lurl <<" ("<<i<<")"<<endl;
    globus_object_t *error;
    if((globus_object_t*)globus_gass_copy_register_url_to_url_autodir(
               &(handles[i].handle),
               (char*)(handles[i].lurl.c_str()),GLOBUS_NULL,
               (char*)(handles[i].rurl.c_str()),&(handles[i].handle_attr),
               &upload_finished_callback,handles+i) != GLOBUS_SUCCESS) {
      /* bad url probably - ???? leave with error ???? */
      odlog(0)<< "Probably bad url or local file: "<<handles[i].rurl<<"  "<<handles[i].lurl<< endl;
      handles[i].active=false;
      handles[i].lurl.resize(0);
      ++next_di;
      globus_mutex_unlock(&(list_lock));
      res=1; break; /* leaving program with error (unrecoverable)*/
    };
    new_upload=true;
    ++next_di;
    globus_mutex_unlock(&(list_lock));
  };
  /* still can be uploads running - must cancel here */
  odlog(1) << "Leaving uploader" << endl;
  leaving=true;
//  globus_mutex_lock(&(list_lock));
//  for(i=0;i<MAX_UPLOADS;i++) {
//    if(handles[i].active)
//      globus_gass_copy_cancel(&(handles[i].handle),
//                   upload_cancel_callback,handles+i);
//  };
//  globus_mutex_unlock(&(list_lock));
  /* just wait a bit */
  sleep(3);
  for(i=0;i<MAX_UPLOADS;i++) {
    globus_gass_copy_handle_destroy(&(handles[i].handle));
    globus_ftp_client_operationattr_destroy(&(handles[i].handle_ftp_attr));
  };
  globus_mutex_destroy(&upload_lock);
  globus_cond_destroy(&upload_cond);
  globus_mutex_destroy(&list_lock);
exit:
  globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
  globus_module_deactivate(GLOBUS_COMMON_MODULE);
#ifdef __UI_LIBRARY__
  if(jobid) {
    if(job_id) { if((*job_id) != jobid) free(jobid); }
    else { free(jobid); };
  };
#else
  if(jobid) free(jobid);
#endif
  return res;
}
