/*
  upload files specified in job.ID.output .
  result: 0 - ok, 1 - unrecoverable error, 2 - potentially recoverable.
*/
#include "../std.h"
#include <list>
#include <globus_common.h>
#include <globus_ftp_client.h>
#include <globus_ftp_client_debug_plugin.h>
#include <globus_gass_copy.h>
#include <globus_object.h>
#include "replica.h"
#include "replica_utils.h"
#include "../files/info_types.h"
#include "../files/info_files.h"
#include "../misc/delete.h"
#include "../misc/url_options.h"
#include "../misc/stringtoint.h"
#include "../misc/log_time.h"
#include "../config/config_map.h"
#include "../cache/cache.h"
#include "check.h"
#include "proxy.h"
#include "convert_url.h"

/* default timeout is 5 minutes per file */
#define DEFAULT_TIMEOUT 1200
/* check for user uploaded files every 10 seconds */
#define CHECK_PERIOD 10
/* maximum number of retries (for all files together) */
#define MAX_RETRIES 5
/* maximum number simultaneous uploads */
#define MAX_UPLOADS 5

void MinusPlus(list<FileData> &from, list<FileDataRC> &to) {
  for(FileData::iterator i=from.begin();i!=from.end();++i) {
    to.push_back(FileDataRC(*i));
  };
//  from.erase();
}

typedef struct {
  globus_gass_copy_handle_t handle;
  globus_ftp_client_operationattr_t ftp_attr;
  globus_gass_copy_attr_t o_attr;
  FileDataRC::iterator di;
  bool active;
  string lurl;     /* local url file:///path */
  string rurl;     /* remote url gsiftp://host:port/path */
  string rcurl;    /* RC url ldap://host:port/DN */
  string path;     /* path to the file == lurl */
  string lfn;      /* url as provided in RSL */
} upload_handle;
upload_handle handles[MAX_UPLOADS];

globus_mutex_t list_lock;
globus_mutex_t upload_lock;
globus_cond_t  upload_cond;
list<FileDataRC> job_files;
JobDescription desc;
JobUser user;
int upload_done=0;
FileDataRC::iterator next_di;
bool leaving = false;
bool rc_failures_happened = false;
const char* cache_dir = NULL;
const char* cache_data_dir = NULL;
const char* cache_link_dir = NULL;
uid_t cache_uid;
gid_t cache_gid;

/* const char* rc_url_head = "rc://"; */

bool register_url(string &url,const string &location,const string &path) {
  /* register file to RC */
  const char *s = url.c_str();
  if(!strncasecmp(s,rc_url_head,strlen(rc_url_head))) {
    string rc_url;
    if(!extract_RC_from_url(url,rc_url)) {
      olog << "URL part to contact Replica Catalog is missing from " << url << endl;
      return false;
    };
    if(rc_url.length() == 0) {
      olog << "URL to contact Replica Catalog not supplied." << endl;
      return false;
    };
    s=url.c_str()+strlen(rc_url_head);
    const char* s_e = strchr(s,'/'); if(s_e == NULL) s_e=s+strlen(s);
    string host(s,s_e-s); /* list of hosts or one host ? */
    if(location.length() != 0) {
      int pos = location.find("://");
      if(pos != string::npos) {
        int pos_=location.find('/',pos+3);
        if(pos == string::npos) pos_=location.length();
        host=location.substr(pos+3,pos_-pos-3);  
      }
      else {
        host=location;
      };
    };
    if((*s_e) == 0) {
      olog<<"Missing file name in url: "<<url<<endl; return false;
    };
    string name(s_e+1);
    cerr<<"Registering file \""<<name<<"\" to Replica Catalog "<<rc_url<<endl; 
    RCFile file(name,path);
    if(!(file.is_valid())) {
      olog<<"Determination of file parameters failed: "<<path<<endl;
      return false;
    };
    RCManager rc(rc_url,"","");
    if(!(rc.is_open())) {
      olog << "Failed accessing Replica Catalog collection: " << rc_url << endl;
      return false;
    };
    if(!(rc.AddFile(file,true))) {
      olog<<"Creation of logical file failed: "<<name<<endl;
      olog<<"Maybe already exist."<<endl;
      return false;
    };
    /* Adding physical location of file. NOTE - location MUST
       already exist on collection */
    if(!(rc.AddFileLocation(file,host,"*",true))) {
      olog<<"Creation of file location failed: "<<name<<" ("<<host<<")"<<endl;
      olog<<"Please, check if this file location/instance exists."<<endl;
      return false;
    };
  };
  return true;
}

int clean_files(list<FileData> &job_files,char* session_dir) {
  string session(session_dir);
  if(delete_all_files(session,(list<FileData>&)job_files,true) != 2) return 0;
  return 1;
}

void upload_cancel_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {
  upload_handle *h = (upload_handle*)arg;
  globus_mutex_lock(&(list_lock));
  h->active=false;
  globus_mutex_unlock(&(list_lock));
};

void upload_finished_callback(void * arg,globus_gass_copy_handle_t * handle,globus_object_t * error) {
  upload_handle *h = (upload_handle*)arg;

olog << "Transfer callback: "<<h->di->lfn<<" ("<<h->di->pfn<<")"<<endl;
  globus_mutex_lock(&(list_lock));
  if(error != GLOBUS_SUCCESS) {
    olog << "Transfer failed: "<<h->di->lfn<<" ("<<h->di->pfn<<")"<<endl;
    char* tmp = globus_object_printable_to_string(error);
    olog << "Error is: "<<tmp<<endl; free(tmp);
    h->lurl.resize(0);
    h->rurl.resize(0);
    h->rcurl.resize(0);
    h->path.resize(0);
    h->active=false;
  }
  else {
    if(!leaving) { /* canceling transfer can appear here as success */
olog << "Uploaded: "<<h->di->lfn<<" ("<<h->di->pfn<<")"<<endl;
      /* make sure erasing does not invalidate current pointer */
      if(next_di == h->di) { next_di=job_files.erase(h->di); }
      else { job_files.erase(h->di); };
      /* save changed list */
      if(!register_url(h->lfn,h->rurl,h->path)) {
        olog << "Failed registering file." << endl;
        rc_failures_happened = true;
      }
      else {
        if(!job_output_write_file(desc,user,(list<FileData>&)job_files)) {
          olog << "Failed writing changed output file." << endl;
        };
      };
    };
    h->lurl.resize(0);
    h->rurl.resize(0);
    h->rcurl.resize(0);
    h->path.resize(0);
    h->active=false;
  };
  globus_mutex_unlock(&(list_lock));
  globus_mutex_lock(&(upload_lock));
  upload_done=1;
  globus_cond_signal(&upload_cond);
  globus_mutex_unlock(&(upload_lock));
// olog << "Transfer callback: leaving" << endl;
}

/* this function will wake up main routine periodically.
   it's main purpose here is to provide timeout for hanged
   connections */
void* timer_func(void* arg) {
  int timeout = (int)arg;
  for(;;) {
    sleep(timeout);
    globus_mutex_lock(&(upload_lock));
    upload_done=2;
    globus_cond_signal(&upload_cond);
    globus_mutex_unlock(&(upload_lock));
  };
};

int main(int argc,char** argv) {
cerr<<"Uploader - first line"<<endl;
  int res=0;
  bool new_upload;
  int retries=0;
  globus_thread_t timer_thread;
  time_t start_time=time(NULL);
  time_t upload_timeout;
  int n_threads = 1;
  LogTime::active = true;
  uid_t file_owner;
  gid_t file_group;

  for(;;) {
    opterr=0;
    char optc=getopt(argc,argv,"+ht:u:U:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"uploader [-t threads] job_id control_directory session_directory"<<endl;
        exit(1);
      }; break;
      case 't': {
        n_threads=atoi(optarg);
        if(n_threads < 1) {
          olog<<"Wrong number of threads"<<endl; exit(1);
        };
      }; break;
      case 'U': {
        unsigned int tuid;
        if(!stringtoint(optarg,tuid)) {
          olog<<"Bad number "<<optarg<<endl; exit(1);
        };
        struct passwd pw_;
        struct passwd *pw;
        char buf[BUFSIZ];
        getpwuid_r(tuid,&pw_,buf,BUFSIZ,&pw);
        if(pw == NULL) {
          olog<<"Wrong user name"<<endl; exit(1);
        };
        file_owner=pw->pw_uid;
        file_group=pw->pw_gid;
        if((getuid() != 0) && (getuid() != file_owner)) {
          olog<<"Specified user can't be handled"<<endl; exit(1);
        };
      }; break;
      case 'u': {
        struct passwd pw_;
        struct passwd *pw;
        char buf[BUFSIZ];
        getpwnam_r(optarg,&pw_,buf,BUFSIZ,&pw);
        if(pw == NULL) {
          olog<<"Wrong user name"<<endl; exit(1);
        };
        file_owner=pw->pw_uid;
        file_group=pw->pw_gid;
        if((getuid() != 0) && (getuid() != file_owner)) {
          olog<<"Specified user can't be handled"<<endl; exit(1);
        };
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };

  char* id = argv[optind+0];
  if(!id) { olog << "Missing job id" << endl; return 1; };
  char* control_dir = argv[optind+1];
  if(!control_dir) { olog << "Missing control directory" << endl; return 1; };
  char* session_dir = argv[optind+2];
  if(!session_dir) { olog << "Missing session directory" << endl; return 1; };
  cache_dir = argv[optind+3];
  if(cache_dir) {
    olog<<"Cache directory is "<<cache_dir<<endl;
    cache_data_dir=argv[optind+4];
    if(!cache_data_dir) {
      cache_data_dir=cache_dir;
    }
    else {
      olog<<"Cache data directory is "<<cache_data_dir<<endl;
      cache_link_dir=argv[optind+5];
      if(cache_link_dir) olog<<"Cache links will go to directory "<<cache_link_dir<<endl;
    };
  };
  cache_uid=getuid(); cache_gid=getgid();
  olog<<"Uploader started"<<endl;
  if(cache_dir) cache_release_url(cache_dir,cache_data_dir,cache_uid,cache_gid,id,true);
  
  prepare_proxy();

  UrlMap url_map;
  if(globus_module_activate(GLOBUS_COMMON_MODULE) != GLOBUS_SUCCESS) {
    olog<<"COMMON module activation failed\n"; return 1;
  };
  if(globus_module_activate(GLOBUS_GASS_COPY_MODULE) != GLOBUS_SUCCESS) {
    olog<<"GASS COPY module activation failed\n";
    globus_module_deactivate(GLOBUS_COMMON_MODULE); return 1;
  };
  if(globus_module_activate(GLOBUS_REPLICA_CATALOG_MODULE) != GLOBUS_SUCCESS) {
    olog<<"REPLICA CATALOG module activation failed\n";
    globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
    globus_module_deactivate(GLOBUS_COMMON_MODULE); return 1;
  };
  desc = JobDescription(id,session_dir);
  uid_t uid = getuid();
  user=JobUser(uid);
  user.SetControlDir(control_dir);
  user.SetSessionRoot(session_dir); /* not needed */
  {
    list<FileData> job_files_;
    if(!job_output_read_file(desc.get_id(),user,job_files_)) {
      olog << "Can't read list of output files" << endl; res=1; goto exit;
    };
    MinusPlus(job_files_,job_files);
    upload_timeout=DEFAULT_TIMEOUT * (job_files.size() + 1);
    if(clean_files(job_files_,session_dir) != 0) {
      olog << "Can't remove unnecessary files" << endl; res=1; goto exit;
    };
  };
  int i;

//
// globus_ftp_client_plugin_t debug_plugin;
// globus_ftp_client_debug_plugin_init(&debug_plugin,stderr,"debug");
//

cerr<<"Init handles"<<endl;
  for(i=0;i<MAX_UPLOADS;i++) {
    if(globus_gass_copy_handle_init(&(handles[i].handle),GLOBUS_NULL) != GLOBUS_SUCCESS) {
      olog << "Failed to init globus handle" << endl; res=2; goto exit;
    };
    globus_ftp_control_parallelism_t ftp_paral;
    globus_ftp_client_operationattr_init(&(handles[i].ftp_attr));
    if(n_threads>1) {
olog<<"Using default parallel upload !"<<endl;
      ftp_paral.fixed.mode=GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
      ftp_paral.fixed.size=n_threads;
      globus_ftp_client_operationattr_set_parallelism(&(handles[i].ftp_attr),&ftp_paral);
    };
    globus_ftp_client_operationattr_set_striped(&(handles[i].ftp_attr),GLOBUS_FALSE);
    globus_ftp_client_operationattr_set_type(&(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_TYPE_IMAGE);
    globus_ftp_client_operationattr_set_mode(&(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
    globus_ftp_client_operationattr_set_data_protection(&(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    globus_ftp_client_operationattr_set_control_protection(&(handles[i].ftp_attr),GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);

//
// globus_ftp_client_handleattr_add_plugin(&(handles[i].ftp_attr),&debug_plugin);
//

    globus_gass_copy_attr_init(&(handles[i].o_attr));
    globus_gass_copy_attr_set_ftp(&(handles[i].o_attr),&(handles[i].ftp_attr));
    handles[i].active=false;
  };
  globus_mutex_init(&list_lock,GLOBUS_NULL);
  globus_mutex_init(&upload_lock,GLOBUS_NULL);
  globus_cond_init(&upload_cond,GLOBUS_NULL);
cerr<<"Init thread"<<endl;
  if(globus_thread_create(&timer_thread,NULL,&timer_func,(void*)CHECK_PERIOD) != 0) {
    olog << "Failed to start new thread" << endl; res=2; goto exit;
  };
  upload_done=0; 
  next_di=job_files.begin();
  new_upload=false;
  for(;;) {
    /* find something to upload */
    globus_mutex_lock(&(list_lock));
    if(next_di == job_files.end()) {
cerr<<"End of list"<<endl;
      if(new_upload) { /* increase retries only if there were new tries */
        new_upload=false;
        /* check retry counter */
        retries++; 
        olog << "Increasing retry counter: " << retries << endl;
        if(retries >= MAX_RETRIES) { 
          olog << "Maximum number of retries exceeded." << endl;
          globus_mutex_unlock(&(list_lock)); res=2; break;
        };
      };   
      /* nothing to upload */
      olog << "Looking for current uploads in progress" << endl;
      /* check for pending uploads */
      for(i=0;i<MAX_UPLOADS;i++) { if(handles[i].active) break; };
      if(i >= MAX_UPLOADS) { /* all uploads completed or out of tries */
        olog << "No current uploads found" << endl;
        globus_mutex_unlock(&(list_lock)); break;
      };
      i=MAX_UPLOADS; /* mark - just wait */
    }
    else {
olog << "Checking: " << *next_di << endl;
      if((*next_di).lfn.find(":") == string::npos) { /* is it lfn */
olog << "Keep locally: " << *next_di << endl;
        ++next_di; globus_mutex_unlock(&(list_lock)); continue;
      };
      /* isn't it already being uploaded */
      for(i=0;i<MAX_UPLOADS;i++) { 
        if((handles[i].active) && (handles[i].di == next_di)) break;
      };
      if(i < MAX_UPLOADS) {
olog << "Being uploaded: " << *next_di << endl;
        ++next_di; globus_mutex_unlock(&(list_lock)); continue;
      };
      /* find unused handle */
      for(i=0;i<MAX_UPLOADS;i++) { if(!handles[i].active) break; };
    };
    if(i >= MAX_UPLOADS) {  /* no handles to use or no new uploads */
olog << "Waiting till upload finishes" << endl;
      /* wait till upload finishes */
      globus_mutex_unlock(&(list_lock));
      globus_mutex_lock(&(upload_lock));
      while(!(upload_done)) {
        globus_cond_wait(&(upload_cond),&(upload_lock));
      };
      upload_done=0;
      globus_mutex_unlock(&(upload_lock));
      globus_mutex_lock(&(list_lock));
      if(next_di==job_files.end())next_di=job_files.begin();
      globus_mutex_unlock(&(list_lock));

      if((time(NULL) - start_time) > upload_timeout) {
        olog << "Upload timeout" << endl; res=2; break;
      };
      continue;
    };
    handles[i].di=next_di; handles[i].active=true;
//    handles[i].lurl=string("file://") + session_dir + '/' + (*next_di).pfn;
    handles[i].lurl=string("file://") + session_dir + (*next_di).pfn;
    /* resolve url: rc -> gsiftp, ... */
    /* handles[i].rurl=next_di->lfn; */
    RCFile ex_file;
    if(!convert_url(handles[i].rurl,*next_di,url_map,true,true,&ex_file)) {
      olog << "URL resolving problem." << endl; res=2;
      globus_mutex_unlock(&(list_lock)); break;
    };
    if(ex_file.is_valid()) {
      olog << "Logical file already exists in RC." << endl; res=2;
      globus_mutex_unlock(&(list_lock)); break;
    };
    handles[i].rcurl = next_di->rc_url;
olog << "Going to start upload: " << (*next_di).lfn << " ("<<handles[i].rurl <<") from " << handles[i].lurl << endl;
    globus_result_t error;
    handles[i].lfn=(*next_di).lfn;
    handles[i].path=string(session_dir) + '/' + (*next_di).pfn;
    globus_gass_copy_attr_t *o_attr_p = NULL;
    /* setting number of download threads */
    string thr_num_s;
    int thr_num = n_threads; /* default value given in options */
    if(get_url_option(handles[i].rurl,"threads",0,thr_num_s) == 0) {
      if(stringtoint(thr_num_s,thr_num)) {
        if(thr_num < 1) thr_num = 1;
        if(thr_num > 10) thr_num = 10;
      }
      else { thr_num = 1; };
    };
    remove_url_options(handles[i].rurl);
    globus_ftp_control_parallelism_t ftp_paral;
    if(thr_num <= 1) {
      ftp_paral.mode=GLOBUS_FTP_CONTROL_PARALLELISM_NONE;
      ftp_paral.base.size=1; /* in case */
    }
    else {
      ftp_paral.mode=GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
      ftp_paral.fixed.size=thr_num;
    };
    globus_ftp_client_operationattr_set_parallelism(&(handles[i].ftp_attr),&ftp_paral);
    if(!strncmp(next_di->lfn.c_str(),"gsiftp://",9)) {
      o_attr_p = &(handles[i].o_attr);
olog<<"Using "<<thr_num<<" threads"<<endl;
    };
    if((error=globus_gass_copy_register_url_to_url_check(
               &(handles[i].handle),
               (char*)(handles[i].lurl.c_str()),GLOBUS_NULL,
               (char*)(handles[i].rurl.c_str()),o_attr_p,
               &upload_finished_callback,handles+i)) != GLOBUS_SUCCESS) {
      globus_object_t *err_obj = globus_error_get(error);
      if((err_obj != GLOBUS_ERROR_LOCAL_FILE_NOT_FOUND) &&
         (error != GLOBUS_ERROR_LOCAL_FILE_NOT_FOUND)) {  
      /* bad url probably - ???? leave with error ???? */
        olog << "Bad url or local file: " << next_di->lfn << " (" << next_di->pfn << ")" << endl;
        char* tmp = globus_object_printable_to_string(globus_error_get(error));
        olog<<"Error is: "<<tmp<<endl; free(tmp);
        handles[i].active=false;
        handles[i].lurl.resize(0);
        ++next_di;
        globus_mutex_unlock(&(list_lock));
        res=1; break; /* leaving program with error (unrecoverable)*/
      }
      else { /* no such local file - this not an error - warning message */
olog << "File does not exist: "<<handles[i].path<<endl;
        handles[i].lurl.resize(0);
        handles[i].rurl.resize(0);
        handles[i].rcurl.resize(0);
        handles[i].path.resize(0);
        handles[i].active=false;
        /* make sure erasing does not invalidate current pointer */
        next_di=job_files.erase(next_di);
        /* save changed list */
        if(!job_output_write_file(desc,user,(list<FileData>&)job_files)) {
          olog << "Failed writing changed output file." << endl;
        };
      };
    }
    else {
olog << "Upload started" << endl;
      new_upload=true;
      ++next_di;
    };
    globus_mutex_unlock(&(list_lock));
  };
  /* still can be uploads running - must cancel here */
olog << "Leaving uploader" << endl;
  leaving=true;
  globus_mutex_lock(&(list_lock));
  for(i=0;i<MAX_UPLOADS;i++) {
    if(handles[i].active)
      globus_gass_copy_cancel(&(handles[i].handle),
                   upload_cancel_callback,handles+i);
  };
  globus_mutex_unlock(&(list_lock));
  /* just wait a bit */
  sleep(3);
//  if(res==0) {
//    if(!job_output_mark_put(desc,user)) {
//      olog << "Failed putting output mark" << endl; res=2;
//    };
//  };
  /* close RC connection */
  /* clean unfinished files here */
  clean_files((list<FileData>&)job_files,session_dir);
  for(i=0;i<MAX_UPLOADS;i++) {
    globus_gass_copy_handle_destroy(&(handles[i].handle));
    globus_ftp_client_operationattr_destroy(&(handles[i].ftp_attr));
//    globus_gass_copy_attr_destroy(&(handles[i].o_attr));
  };
  globus_mutex_destroy(&upload_lock);
  globus_cond_destroy(&upload_cond);
  globus_mutex_destroy(&list_lock);
exit:
  if((res == 0)  &&  (rc_failures_happened)) res=2; /* this will make user 
                                                       know about failures */
  remove_proxy();
olog<<"Beeing BAD";
  exit(res);
  globus_module_deactivate(GLOBUS_REPLICA_CATALOG_MODULE);
  globus_module_deactivate(GLOBUS_GASS_COPY_MODULE);
  globus_module_deactivate(GLOBUS_COMMON_MODULE);
  return res;
}
