#include "std.h"

#include "MdsQuery.h"
#include "Xrsl.h"
#include "RemoteFile.h"
#include "Target.h"


Target::Target (Cluster * c, Queue * q) : cluster (c),
					  queue (q),
					  xrsl (NULL),
					  neededcachesize (UNDEFINED),
					  neededsessdirsize (UNDEFINED),
					  remotesize (UNDEFINED),
					  localsize (UNDEFINED) {};

bool Target::operator! () const { return !xrsl; }


int Target::CalculateNeededFileSizes () {

  vector <string> inputfiles;
  if (xrsl.GetInputFiles (inputfiles)) return 1;
  string defaultrc;
  if (xrsl.GetRc (defaultrc)) return 1;

  for (vector <string>::iterator vsi = inputfiles.begin();
       vsi != inputfiles.end(); vsi++) {

    string url = *vsi;
    string optstring;

    int pos = url.find ("://");
    if (pos != string::npos) {
      pos += 3;
      int atpos = url.find ('@', pos);
      if (atpos != string::npos) {
	int slashpos = url.find ('/', pos);
	if (slashpos == string::npos || atpos < slashpos) pos = atpos + 1;
      }
      int semicpos = url.find (';', pos);
      if (semicpos != string::npos) {
	int slashpos = url.find ('/', pos);
	if (slashpos == string::npos) {
	  optstring = url.substr (semicpos + 1);
	  url.erase (semicpos);
	}
	else if (slashpos > semicpos) {
	  optstring = url.substr (semicpos + 1, slashpos - semicpos - 1);
	  url.erase (semicpos, slashpos - semicpos);
	}
      }
    }

    if (url.substr (0, 5) == "rc://") {

      string locstring;
      string rc;

      int pos = 5;
      int slashpos = url.find ('/', pos);
      int atpos = url.find ('@', pos);
      if (slashpos == string::npos) {
	cerr << "Can not parse rc:// URL" << endl;
	return 1;
      }
      if (atpos != string::npos && atpos < slashpos) {
	locstring = url.substr (pos, atpos - pos);
	pos = atpos + 1;
      }
      else if (strcasecmp (url.substr (slashpos + 1, 3).c_str(), "lc=") != 0) {
	locstring = url.substr (pos, slashpos - pos);
	pos = slashpos;
      }
      if (pos == slashpos) {
	rc = defaultrc;
	if (rc.empty()) {
	  cerr << "XRSL attribute \"replicacollection\" missing" << endl;
	  return 1;
	}
      }
      else {
	slashpos++;
	slashpos = url.find ('/', slashpos);
	if (slashpos == string::npos) {
	  cerr << "Can not parse rc:// URL" << endl;	
	  return 1;
	}
	rc = "ldap://" + url.substr (pos, slashpos - pos);
      }
      string file = url.substr (slashpos + 1);
      query.AddFile (new RcFile (rc, file, locstring, optstring));
    }

    else if (url.substr (0, 6) == "ftp://" ||
	     url.substr (0, 9) == "gsiftp://") {
      query.AddFile (new FtpFile (url, optstring));
    }

    else if (url.substr (0, 7) == "http://" ||
	     url.substr (0, 8) == "https://") {
      query.AddFile (new HttpFile (url, optstring));
    }

    else if (url.substr (0, 7) == "file://" ||
	     url.find ("://") == string::npos) {
      if (url.substr (0, 7) == "file://")
	url = url.substr (7);
      else if (url[0] != '/') {
	char buffer[PATH_MAX];
	getcwd (buffer, PATH_MAX);
	url = (string) buffer + '/' + url;
      }
      query.AddFile (new LocalFile (url, optstring));
    }

  }

  if (query.AddSizes (cluster, &neededcachesize, &neededsessdirsize,
		      &remotesize, &localsize)) return 1;

  long long int disk;
  if (xrsl.GetDisk (&disk)) return 1;
  if (disk != UNDEFINED) neededsessdirsize += disk;

  return 0;
}


long long int Target::GetNeededCacheSize () const {

  return neededcachesize;
}


long long int Target::GetNeededSessDirSize () const {

  return neededsessdirsize;
}


long long int Target::GetNeededTotalSize () const {

  if (neededcachesize == UNDEFINED || neededsessdirsize == UNDEFINED)
    return UNDEFINED;
  return neededcachesize + neededsessdirsize;
}


long long int Target::GetRemoteSize () const {

  return remotesize;
}


long long int Target::GetLocalSize () const {

  return localsize;
}


int Target::Accept () {

  int time;
  if (xrsl.GetCpuTime (&time)) return 1;
  if (time == UNDEFINED) time = queue->GetDefaultCpuTime();

  int count;
  if (xrsl.GetCount (&count)) return 1;
  if (count == UNDEFINED) count = 1;

  queue->Accept (time, count);

  if (cluster->GetCacheFree() != UNDEFINED)
    cluster->ClaimCache (neededcachesize);
  else {
    cluster->ClaimSessionDir (neededcachesize);
    queue->ClaimUserDiskSpace (neededcachesize);
  }

  cluster->ClaimSessionDir (neededsessdirsize);
  queue->ClaimUserDiskSpace (neededsessdirsize);

  if (query.RegisterCachedFiles (cluster)) return 1;

  return 0;
}


int Target::GetCpuNumber () const {

  int ncpus = UNDEFINED;

  if (queue->GetAssignedCpuNumber() != UNDEFINED &&
      (ncpus == UNDEFINED || queue->GetAssignedCpuNumber() < ncpus))
    ncpus = queue->GetAssignedCpuNumber();
  if (cluster->GetTotalCpus() != UNDEFINED &&
      (ncpus == UNDEFINED || cluster->GetTotalCpus() < ncpus))
    ncpus = cluster->GetTotalCpus();
  return (ncpus == UNDEFINED) ? 0 : ncpus;
}


int Target::GetQueued () const {

  int queued = UNDEFINED;

  if (queue->GetQueued() != UNDEFINED &&
      (queued == UNDEFINED || queue->GetQueued() > queued))
    queued = queue->GetQueued();
  if (cluster->GetQueued() != UNDEFINED &&
      (queued == UNDEFINED || cluster->GetQueued() > queued))
    queued = cluster->GetQueued();
  return (queued == UNDEFINED) ? INT_MAX : queued;
}
