#include "std.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <getopt.h>
#include <unistd.h>
#include <stdio.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"

#define __UI_LIBRARY__
#include "ui_downloader.h"

#ifdef __CYGWIN__
#define getopt_long_only getopt_long
#endif


void ngcatusage (char * progname) {

  cout << "Usage: " << progname << " [options] [job ...]" << endl;
  cout << endl;
  cout << "Options:" << endl;
  cout << "  -a, -all                     all jobs" << endl;
  cout << "  -i, -joblist   filename      file containing a list of jobids" << endl;
  cout << "  -o, -stdout                  show stdout of the job (default)" << endl;
  cout << "  -e, -stderr                  show stderr of the job" << endl;
  cout << "  -l, -gmlog                   show the grid manager's error log of the job" << endl;
  cout << "  -t, -timeout   time          timeout for MDS queries in seconds (default " << DEFAULT_TIMEOUT << ")" << endl;
  cout << "  -d, -debug     debuglevel    0 = none, 1 = some, 2 = more, 3 = a lot" << endl;
  cout << "  -x, -anonymous               use anonymous bind for MDS queries (default)" << endl;
  cout << "  -X, -gsi                     use gsi-gssapi bind for MDS queries" << endl;
  cout << "  -v, -version                 print version information" << endl;
  cout << "  -h, -help                    print this help" << endl;
  cout << endl;
  cout << "Arguments:" << endl;
  cout << "  job ...                      list of jobids and/or jobnames" << endl;
  cout << endl;
  cout << "Default values for the MDS timeout or the debuglevel can be set by defining" << endl;
  cout << "the environment variables NGTIMEOUT or NGDEBUG, or if these are undefined by" << endl;
  cout << "defining NGTIMEOUT or NGDEBUG in $(HOME)/.ngrc" << endl;
}


int ngcat (int argc, char ** argv) {

  // options

  bool all = false;
  vector <string> joblists;
  bool stderr = false;
  bool gmlog = false;
  int timeout = UNDEFINED;
  int debug = UNDEFINED;
  bool anonymous = true;

  // parsing options

  char optstring[] = "ai:oelt:d:xXvh";
  struct option longoptions[] = {
    {"all", 0, 0, 'a'},
    {"joblist", 1, 0, 'i'},
    {"stdout", 0, 0, 'o'},
    {"stderr", 0, 0, 'e'},
    {"gmlog", 0, 0, 'l'},
    {"timeout", 1, 0, 't'},
    {"debug", 1, 0, 'd'},
    {"anonymous", 0, 0, 'x'},
    {"gsi", 0, 0, 'X'},
    {"version", 0, 0, 'v'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  int opt;

  while (opt != -1) {

    opt = getopt_long_only (argc, argv, optstring, longoptions, NULL);

    switch (opt) {
    case -1:
      break;
    case 'a':
      all = true;
      break;
    case 'i':
      joblists.push_back (optarg);
      break;
    case 'o':
      stderr = false;
      gmlog = false;
      break;
    case 'e':
      stderr = true;
      gmlog = false;
      break;
    case 'l':
      stderr = false;
      gmlog = true;
      break;
    case 't':
      timeout = atoi (optarg);
      break;
    case 'd':
      debug = atoi (optarg);
      break;
    case 'x':
      anonymous = true;
      break;
    case 'X':
      anonymous = false;
      break;
    case 'v':
      cout << argv[0] << ": version " << VERSION << endl;
      return 0;
    case 'h':
      ngcatusage (argv[0]);
      return 0;
    default:
      cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
      return 1;
    }
  }

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) return 1;

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft() / 60) << endl;
  }

  // finding the information

  vector <string> jobids = GetJobIDs (argc, argv, optind, joblists, all);

  if (all && jobids.empty()) {
    cout << "No jobs" << endl;
    return 0;
  }

  if (jobids.empty()) {
    cerr << argv[0] << ": no valid jobids given" << endl;
    cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
    return 1;
  }

  ClusterList cl;

  int error = 0;

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {
    int pos0 = 0;
    if (vsi->substr (0, 9) == "gsiftp://") pos0 = 9;
    int pos = vsi->find_first_of (":/", pos0);
    if (pos == string::npos) {
      cerr << argv[0] << ": invalid jobid: " << *vsi << endl;
      error = 1;
      continue;
    }
    string clustername = vsi->substr (pos0, pos - pos0);
    bool found = false;
    for (ClusterList::iterator cli = cl.begin(); cli != cl.end(); cli++)
      if (clustername == cli->GetName()) {found = true; break;}
    if (!found) cl.push_back (clustername);
  }

  if (cl.empty()) {
    cerr << argv[0] << ": no valid jobids given" << endl;
    return 1;
  }

  FindClusterInfo (cl, user.GetSN(), anonymous, timeout, debug);

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {

    Cluster * c = NULL;
    string file;
    string status;
    bool found = false;

    for (ClusterList::iterator cli = cl.begin(); cli != cl.end(); cli++) {
      for (QueueList::iterator qli = cli->queues.begin();
	   qli != cli->queues.end(); qli++) {
	for (JobList::iterator jli = qli->jobs.begin();
	     jli != qli->jobs.end(); jli++) {
	  if (*vsi == jli->GetId()) {
	    c = &*cli;
	    status = jli->GetStatus();
	    if (stderr)
	      file = jli->GetStderr();
	    else if (gmlog)
	      file = jli->GetGmlog();
	    else
	      file = jli->GetStdout();
	    found = true; break;
	  }
	}
	if (found) break;
      }
      if (found) break;
    }

    if (!found) {
      cerr << argv[0] << ": no job with jobid " << *vsi << " found" << endl;
      error = 1;
    }
    else if (status == "ACCEPTED" || status == "PREPARING") {
      cerr << argv[0] << ": job " << *vsi << " has not started yet" << endl;
      error = 1;
    }
    else if (file.empty()) {
      cerr << argv[0] << ": cannot determine name of "
	   << ((stderr) ? "stderr" :
	       ((gmlog) ? "grid manager error log" : "stdout"))
	   << " for " << *vsi << endl;
      error = 1;
    }
    else {
      int pos = vsi->rfind ('/');
      if (pos == string::npos) {
	cerr << argv[0] << ": invalid jobid: " << *vsi << endl;
	error = 1;
	continue;
      }
      string shortid = vsi->substr (pos + 1);

      vector <string> filenames;

      if (gmlog) file += "/errors";
      filenames.push_back (file);

      int err = ui_downloader ((c->GetContact() + '/' + shortid).c_str(),
			       false, ("/tmp/" + shortid).c_str(),
			       filenames, true, false, debug);

      if (err) {
	cerr << argv[0] << ": Download incomplete." << endl;
	error = 1;
	continue;
      }

      system (("cat /tmp/" + shortid + '/' + file).c_str());
      remove (("/tmp/" + shortid + '/' + file).c_str());
      pos = file.rfind ('/');
      while (pos != string::npos) {
	rmdir (("/tmp/" + shortid + '/' + file.substr (0, pos)).c_str());
	pos = file.rfind ('/', pos - 1);
      }
      rmdir (("/tmp/" + shortid).c_str());
    }
  }

  return error;
}
