#include "std.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <getopt.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"
#include "Giis.h"

#ifdef __CYGWIN__
#define getopt_long_only getopt_long
#endif


void ngstatusage (char * progname) {

  cout << "Usage: " << progname << " [options] [job ...]" << endl;
  cout << endl;
  cout << "Options:" << endl;
  cout << "  -g, -giisurl   url           url to the central giis" << endl;
  cout << "  -a, -all                     all jobs" << endl;
  cout << "  -i, -joblist   filename      file containing a list of jobids" << endl;
  cout << "  -c, -clusters                show information about clusters" << endl;
  cout << "  -l, -long                    long format (more information)" << endl;
  cout << "  -t, -timeout   time          timeout for MDS queries in seconds (default " << DEFAULT_TIMEOUT << ")" << endl;
  cout << "  -d, -debug     debuglevel    0 = none, 1 = some, 2 = more, 3 = a lot" << endl;
  cout << "  -x, -anonymous               use anonymous bind for MDS queries (default)" << endl;
  cout << "  -X, -gsi                     use gsi-gssapi bind for MDS queries" << endl;
  cout << "  -v, -version                 print version information" << endl;
  cout << "  -h, -help                    print this help" << endl;
  cout << endl;
  cout << "Arguments:" << endl;
  cout << "  job ...                      list of jobids and/or jobnames" << endl;
  cout << endl;
  cout << "The url to the giis has the format ldap://hostname[:port]/basedn" << endl;
  cout << "If the port is omitted the default 2135 is used" << endl;
  cout << "If the -g or -giisurl option is not given a list of urls is read from the" << endl;
  cout << "first existing file in the following list:" << endl;
  cout << "  1. $(HOME)/.nggiislist" << endl;
  cout << "  2. $(NORDUGRID_LOCATION)/etc/giislist" << endl;
  cout << "        or /opt/nordugrid/etc/giislist if $(NORDUGRID_LOCATION) is undefined" << endl;
  cout << "  3. /etc/giislist" << endl;
  cout << "Default values for the MDS timeout or the debuglevel can be set by defining" << endl;
  cout << "the environment variables NGTIMEOUT or NGDEBUG, or if these are undefined by" << endl;
  cout << "defining NGTIMEOUT or NGDEBUG in $(HOME)/.ngrc" << endl;
}


int ngstat (int argc, char ** argv) {

  // options

  vector <Giis> giislist;
  bool all = false;
  vector <string> joblists;
  bool clusters = false;
  bool longlist = false;
  int timeout = UNDEFINED;
  int debug = UNDEFINED;
  bool anonymous = true;

  // parsing options

  char optstring[] = "g:ai:clt:d:xXvh";
  struct option longoptions[] = {
    {"giisurl", 1, 0, 'g'},
    {"all", 0, 0, 'a'},
    {"joblist", 1, 0, 'i'},
    {"cluster", 0, 0, 'c'},
    {"long", 0, 0, 'l'},
    {"timeout", 1, 0, 't'},
    {"debug", 1, 0, 'd'},
    {"anonymous", 0, 0, 'x'},
    {"gsi", 0, 0, 'X'},
    {"version", 0, 0, 'v'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  int opt;

  while (opt != -1) {

    opt = getopt_long_only (argc, argv, optstring, longoptions, NULL);

    switch (opt) {
    case -1:
      break;
    case 'g':
      {
	Giis giis (optarg);
	if (!giis) {
	  cerr << "Syntax error in giisurl argument: " << optarg << endl;
	  return 1;
	}
	else
	  giislist.push_back (giis);
      }
      break;
    case 'a':
      all = true;
      break;
    case 'i':
      joblists.push_back (optarg);
      break;
    case 'c':
      clusters = true;
      break;
    case 'l':
      longlist = true;
      break;
    case 't':
      timeout = atoi (optarg);
      break;
    case 'd':
      debug = atoi (optarg);
      break;
    case 'x':
      anonymous = true;
      break;
    case 'X':
      anonymous = false;
      break;
    case 'v':
      cout << argv[0] << ": version " << VERSION << endl;
      return 0;
    case 'h':
      ngstatusage (argv[0]);
      return 0;
    default:
      cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
      return 1;
    }
  }

  if ((all || !joblists.empty() || argc > optind) && clusters) {
    cerr << argv[0] << ": incompatible options" << endl;
    cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
    return 1;
  }

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) return 1;

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft() / 60) << endl;
  }

  // finding the information

  if (clusters) {

    if (giislist.empty()) if (GetGiises (giislist)) return 1;

    ClusterList cl = FindClusters (giislist, user.GetSN(), anonymous,
				   timeout, debug);
    if (cl.empty()) {
      cerr << argv[0] << ": could not retrieve cluster list from giis" << endl;
      return 1;
    }

    FindClusterInfo (cl, user.GetSN(), anonymous, timeout, debug);

    for (ClusterList::iterator cli = cl.begin(); cli != cl.end(); cli++) {
      if (longlist)
	cli->Print();
      else
	cli->PrintShort();
      for (QueueList::iterator qli = cli->queues.begin();
	   qli != cli->queues.end(); qli++) {
	if (longlist)
	  qli->Print();
	else
	  qli->PrintShort();
      }
      cout << endl;
    }

    return 0;
  }

  else {

    vector <string> jobids = GetJobIDs (argc, argv, optind, joblists, all);

    if (all && jobids.empty()) {
      cout << "No jobs" << endl;
      return 0;
    }

    if (jobids.empty()) {
      cerr << argv[0] << ": no valid jobids given" << endl;
      cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
      return 1;
    }

    ClusterList cl;

    int error = 0;

    for (vector <string>::iterator vsi = jobids.begin();
	 vsi != jobids.end(); vsi++) {
      int pos0 = 0;
      if (vsi->substr (0, 9) == "gsiftp://") pos0 = 9;
      int pos = vsi->find_first_of (":/", pos0);
      if (pos == string::npos) {
	cerr << argv[0] << ": invalid jobid: " << *vsi << endl;
	error = 1;
	continue;
      }
      string clustername = vsi->substr (pos0, pos - pos0);
      bool found = false;
      for (ClusterList::iterator cli = cl.begin(); cli != cl.end(); cli++)
	if (clustername == cli->GetName()) {found = true; break;}
      if (!found) cl.push_back (clustername);
    }

    if (cl.empty()) {
      cerr << argv[0] << ": no valid jobids given" << endl;
      return 1;
    }

    FindClusterInfo (cl, user.GetSN(), anonymous, timeout, debug);

    for (vector <string>::iterator vsi = jobids.begin();
	 vsi != jobids.end(); vsi++) {
      bool found = false;
      for (ClusterList::iterator cli = cl.begin(); cli != cl.end(); cli++) {
	for (QueueList::iterator qli = cli->queues.begin();
	     qli != cli->queues.end(); qli++) {
	  for (JobList::iterator jli = qli->jobs.begin();
	       jli != qli->jobs.end(); jli++) {
	    if (*vsi == jli->GetId()) {
	      if (longlist)
		jli->Print();
	      else
		jli->PrintShort();
	      found = true; break;
	    }
	  }
	  if (found) break;
	}
	if (found) break;
      }
      if (!found) {
	cerr << argv[0] << ": no job with jobid " << *vsi << " found" << endl;
	error = 1;
      }
    }

    return error;
  }
}
