#include "std.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <getopt.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"
#include "Xrsl.h"
#include "Giis.h"
#include "Target.h"

#define __UI_LIBRARY__
#include "ui_uploader.h"

#ifdef __CYGWIN__
#define getopt_long_only getopt_long
#endif


void ngsubusage (char * progname) {

  cout << "Usage: " << progname << " [options] [xrsl ...]" << endl;
  cout << endl;
  cout << "Options:" << endl;
  cout << "  -g, -giisurl   url           url to the central giis" << endl;
  cout << "  -c, -cluster   [-]name       explicity select or reject a specific cluster" << endl;
  cout << "  -f, -file      filename      xrslfile describing the job to be submitted" << endl;
  cout << "  -o, -joblist   filename      file where the jobids will be stored" << endl;
  cout << "  -t, -timeout   time          timeout for MDS queries in seconds (default " << DEFAULT_TIMEOUT << ")" << endl;
  cout << "  -d, -debug     debuglevel    0 = none, 1 = some, 2 = more, 3 = a lot" << endl;
  cout << "  -x, -anonymous               use anonymous bind for MDS queries (default)" << endl;
  cout << "  -X, -gsi                     use gsi-gssapi bind for MDS queries" << endl;
  cout << "      -dryrun                  add dryrun option to the xrsl" << endl;
  cout << "      -dumpxrsl                do not submit - dump transformed xrsl to stdout" << endl;
  cout << "  -v, -version                 print version information" << endl;
  cout << "  -h, -help                    print this help" << endl;
  cout << endl;
  cout << "Arguments:" << endl;
  cout << "  xrsl ...                     xrslstrings describing the jobs to be submitted" << endl;
  cout << endl;
  cout << "The url to the giis has the format ldap://hostname[:port]/basedn" << endl;
  cout << "If the port is omitted the default 2135 is used" << endl;
  cout << "If the -g or -giisurl option is not given a list of urls is read from the" << endl;
  cout << "first existing file in the following list:" << endl;
  cout << "  1. $(HOME)/.nggiislist" << endl;
  cout << "  2. $(NORDUGRID_LOCATION)/etc/giislist" << endl;
  cout << "        or /opt/nordugrid/etc/giislist if $(NORDUGRID_LOCATION) is undefined" << endl;
  cout << "  3. /etc/giislist" << endl;
  cout << "Default values for the MDS timeout or the debuglevel can be set by defining" << endl;
  cout << "the environment variables NGTIMEOUT or NGDEBUG, or if these are undefined by" << endl;
  cout << "defining NGTIMEOUT or NGDEBUG in $(HOME)/.ngrc" << endl;
}


int ngsub (int argc, char ** argv) {

  // options

  vector <Giis> giislist;
  vector <string> clusterselect;
  vector <string> clusterreject;
  vector <string> xrslfiles;
  string joblist;
  int timeout = UNDEFINED;
  int debug = UNDEFINED;
  bool anonymous = true;
  bool dryrun = false;
  bool dumpxrsl = false;
  vector <string> xrslstrings;

  // parsing options

  char optstring[] = "g:c:f:o:t:d:xXvh";
  struct option longoptions[] = {
    {"giisurl", 1, 0, 'g'},
    {"cluster", 1, 0, 'c'},
    {"file", 1, 0, 'f'},
    {"joblist", 1, 0, 'o'},
    {"timeout", 1, 0, 't'},
    {"debug", 1, 0, 'd'},
    {"anonymous", 0, 0, 'x'},
    {"gsi", 0, 0, 'X'},
    {"dryrun", 0, 0, 'D'},
    {"dumpxrsl", 0, 0, 'p'},
    {"version", 0, 0, 'v'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  int opt;

  while (opt != -1) {

    opt = getopt_long_only (argc, argv, optstring, longoptions, NULL);

    switch (opt) {
    case -1:
      break;
    case 'g':
      {
	Giis giis (optarg);
	if (!giis) {
	  cerr << "Syntax error in giisurl argument: " << optarg << endl;
	  return 1;
	}
	else
	  giislist.push_back (giis);
      }
      break;
    case 'c':
      if (optarg[0] == '-')
	clusterreject.push_back (&optarg[1]);
      else if (optarg[0] == '+')
	clusterselect.push_back (&optarg[1]);
      else
	clusterselect.push_back (optarg);
      break;
    case 'f':
      xrslfiles.push_back (optarg);
      break;
    case 'o':
      joblist = optarg;
      break;
    case 't':
      timeout = atoi (optarg);
      break;
    case 'd':
      debug = atoi (optarg);
      break;
    case 'x':
      anonymous = true;
      break;
    case 'X':
      anonymous = false;
      break;
    case 'D':
      dryrun = true;
      break;
    case 'p':
      dumpxrsl = true;
      break;
    case 'v':
      cout << argv[0] << ": version " << VERSION << endl;
      return 0;
    case 'h':
      ngsubusage (argv[0]);
      return 0;
    default:
      cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
      return 1;
    }
  }

  while (argc > optind) xrslstrings.push_back (argv [optind++]);

  if (xrslfiles.empty() && xrslstrings.empty()) {
    cerr << argv[0] << ": no xrsl input specified" << endl;
    cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
    return 1;
  }

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) return 1;

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft() / 60) << endl;
  }

  // parse xrsl

  vector <Xrsl> xrsllist;

  for (vector <string>::iterator vsi = xrslfiles.begin();
       vsi != xrslfiles.end(); vsi++) {
    Xrsl xrsl = Xrsl (*vsi, 0);
    if (!xrsl) {
      cerr << "The XRSL file \"" << *vsi << "\" could not be parsed" << endl;
      return 1;
    }
    bool done = false;
    for (int i = 0; !done; i++) {
      Xrsl axrsl = xrsl[i];
      if (!axrsl)
	done = true;
      else
	xrsllist.push_back (axrsl);
    }
  }

  for (vector <string>::iterator vsi = xrslstrings.begin();
       vsi != xrslstrings.end(); vsi++) {
    Xrsl xrsl = *vsi;
    if (!xrsl) {
      cerr << "The XRSL string \"" << *vsi << "\" could not be parsed" << endl;
      return 1;
    }
    bool done = false;
    for (int i = 0; !done; i++) {
      Xrsl axrsl = xrsl[i];
      if (!axrsl)
	done = true;
      else
	xrsllist.push_back (axrsl);
    }
  }

  for (vector <Xrsl>::iterator vxi = xrsllist.begin();
       vxi != xrsllist.end(); vxi++) {

    if (!clusterselect.empty())
      vxi->AddSimpleRelationSet ("cluster", clusterselect, true);
    if (!clusterreject.empty())
      vxi->AddSimpleRelationSet ("cluster", clusterreject, false);

    if (debug > 1) {
      cout << "Parsed input XRSL:" << endl;
      vxi->Print();
    }

    vxi->Eval();

    if (vxi->Check()) return 1;
  }

  // find available clusters

  if (giislist.empty()) if (GetGiises (giislist)) return 1;

  ClusterList cl = FindClusters (giislist, user.GetSN(), anonymous,
				 timeout, debug);
  if (cl.empty()) {
    cerr << argv[0] << ": could not retrieve cluster list from giis" << endl;
    return 1;
  }

  FindClusterInfo (cl, user.GetSN(), anonymous, timeout, debug);

  // do brokering

  srand (time (NULL));

  for (vector <Xrsl>::iterator vxi = xrsllist.begin();
       vxi != xrsllist.end(); vxi++) {

    vector <Target> targetlist;

    for (ClusterList::iterator cli = cl.begin(); cli != cl.end(); cli++) {
      for (QueueList::iterator qli = cli->queues.begin();
	   qli != cli->queues.end(); qli++) {

	Target target (&*cli, &*qli);

	if (debug) {
	  cout << "Cluster: " << target.cluster->GetLongName() << endl;
	  cout << "Queue: " << target.queue->GetName() << endl;
	}

	if (target.queue->GetStatus() != "active") {
	  if (debug) cout << "Queue rejected due to status: "
			  << target.queue->GetStatus() << endl;
	  continue;
	}

	if (target.queue->GetUserFreeCpus() == UNDEFINED) {
	  if (debug) cout << "Queue rejected because user not authorized"
			  << endl;
	  continue;
	}

	if (target.queue->GetMaxQueuable() != UNDEFINED &&
	    target.queue->GetQueued() >= target.queue->GetMaxQueuable()) {
	  if (debug) cout << "Queue rejected because the maximum number "
			  << "of queued jobs is reached" << endl;
	  continue;
	}

	string failattr;
	if (vxi->Test (target, failattr)) return 1;

	if (!target) {
	  if (debug) cout << "Queue rejected because it does not match "
			  << "the XRSL specification (" << failattr << ")"
			  << endl;
	  continue;
	}

	if (target.CalculateNeededFileSizes()) return 1;

	if (target.GetRemoteSize() == UNDEFINED) {
	  if (debug) cout << "Queue rejected because no valid location was "
			  << "found for the inputfiles" << endl;
	  continue;
	}

	if (debug > 1) cout << "Total size of remote files: "
			    << target.GetRemoteSize() << " bytes" << endl;
	if (debug > 1) cout << "Total size of local files: "
			    << target.GetLocalSize() << " bytes" << endl;

	if (target.cluster->GetCacheFree() == UNDEFINED) {
	  if (debug > 1) cout << "Total size of files: "
			      << target.GetNeededTotalSize()
			      << " bytes" << endl;

	  if (target.GetNeededTotalSize() > target.queue->GetUserDiskSpace()) {
	    if (debug) cout << "Queue rejected due to insufficient disk space"
			    << endl;
	    continue;
	  }
	}
	else {
	  if (debug > 1) cout << "Total size of files to be cached: "
			      << target.GetNeededCacheSize()
			      << " bytes" << endl;
	  if (debug > 1) cout << "Total size of files in session directory: "
			      << target.GetNeededSessDirSize()
			      << " bytes" << endl;

	  if (target.GetNeededCacheSize() > target.cluster->GetCacheFree()) {
	    if (debug) cout << "Queue rejected due to insufficient cache space"
			    << endl;
	    continue;
	  }
	  if (target.GetNeededSessDirSize() >
	      target.queue->GetUserDiskSpace()) {
	    if (debug) cout << "Queue rejected due to insufficient disk space"
			    << endl;
	    continue;
	  }
	}

	targetlist.push_back (target);
	if (debug) cout << "Queue accepted as possible submission target"
			<< endl;

      }
    }

    bool submitted = false;
    while (!submitted) {

      int freecpus = 0;
      long long int minremotesize = UNDEFINED;
      long long int minlocalsize = UNDEFINED;

      for (vector <Target>::iterator vti = targetlist.begin();
	   vti != targetlist.end(); vti++) {

	if (vti->queue->GetMaxRunning() != UNDEFINED &&
	    vti->queue->GetRunning() >= vti->queue->GetMaxRunning()) continue;

	int time;
	if (vti->xrsl.GetCpuTime (&time)) return 1;
	if (time == UNDEFINED) time = vti->queue->GetDefaultCpuTime();
	int userfreecpus = vti->queue->GetUserFreeCpus (time);

	int count;
	if (vti->xrsl.GetCount (&count)) return 1;
	if (count == UNDEFINED) count = 1;

	if (userfreecpus < count) continue;

	if (minremotesize == UNDEFINED ||
	    vti->GetRemoteSize() < minremotesize) {
	  minremotesize = vti->GetRemoteSize();
	  minlocalsize = vti->GetLocalSize();
	  freecpus = userfreecpus;
	}
	else if (vti->GetRemoteSize() == minremotesize) {
	  if (vti->GetLocalSize() < minlocalsize) {
	    minlocalsize = vti->GetLocalSize();
	    freecpus = userfreecpus;
	  }
	  else
	    freecpus += userfreecpus;
	}
      }

      vector <Target>::iterator target = targetlist.end();

      if (freecpus > 0) {
	float rn = rand() / (RAND_MAX + 1.0) * freecpus;
	int counter = 0;
	bool found = false;

	for (vector <Target>::iterator vti = targetlist.begin();
	     !found && (vti != targetlist.end()); vti++) {

	  if (vti->queue->GetMaxRunning() != UNDEFINED &&
	      vti->queue->GetRunning() >= vti->queue->GetMaxRunning())
	    continue;

	  int time;
	  if (vti->xrsl.GetCpuTime (&time)) return 1;
	  if (time == UNDEFINED) time = vti->queue->GetDefaultCpuTime();
	  int userfreecpus = vti->queue->GetUserFreeCpus (time);

	  int count;
	  if (vti->xrsl.GetCount (&count)) return 1;
	  if (count == UNDEFINED) count = 1;

	  if (userfreecpus < count) continue;

	  if (vti->GetRemoteSize() == minremotesize &&
	      vti->GetLocalSize() == minlocalsize) {
	    counter += userfreecpus;
	    if (rn < counter) {
	      target = vti;
	      found = true;
	    }
	  }
	}
      }

      else {
	for (vector <Target>::iterator vti = targetlist.begin();
	     vti != targetlist.end(); vti++) {

	  int count;
	  if (vti->xrsl.GetCount (&count)) return 1;
	  if (count == UNDEFINED) count = 1;

	  if (vti->GetCpuNumber() < count) continue;

	  if (target == targetlist.end() ||
	      (float) vti->GetQueued() / vti->GetCpuNumber() <
	      (float) target->GetQueued() / target->GetCpuNumber())
	    target = vti;
	}
      }

      if (target == targetlist.end()) {
	cerr << "No cluster found" << endl;
	return 1;
      }

      if (debug) {
	cout << target->cluster->GetLongName() << " selected" << endl;
	cout << "queue " << target->queue->GetName() << " selected" << endl;
      }


      target->xrsl.AddSimpleRelation ("queue", target->queue->GetName());
      target->xrsl.AddSimpleRelation ("action", "request");
      target->xrsl.AddSimpleRelation ("savestate", "yes");
      if (dryrun) target->xrsl.AddSimpleRelation ("dryrun", "dryrun");
      if (target->xrsl.FixJoin()) return 1;
      if (target->xrsl.FixInOut ("executable", "inputfiles")) return 1;
      if (target->xrsl.FixInOut ("stdin", "inputfiles")) return 1;
      if (target->xrsl.FixInOut ("stdout", "outputfiles")) return 1;
      if (target->xrsl.FixInOut ("stderr", "outputfiles")) return 1;
      if (target->xrsl.FixInOut ("gmlog", "outputfiles")) return 1;
      if (target->xrsl.FixExec()) return 1;
      if (target->xrsl.FixStdin()) return 1;
      if (target->xrsl.FixTime ("starttime")) return 1;
      if (target->xrsl.FixPeriod ("cputime")) return 1;
      if (target->xrsl.FixPeriod ("lifetime")) return 1;

      char hostname[1024];
      globus_libc_gethostname (hostname, 1024);
      target->xrsl.AddSimpleRelation ("hostname", hostname);

      string jobname;
      if (target->xrsl.GetJobName (jobname)) return 1;

      vector <string> filenames;
      if (target->xrsl.PrepareUpload (filenames)) return 1;

      if (target->xrsl.Collect (*target, "runtimeenvironment")) return 1;
      if (target->xrsl.Collect (*target, "middleware")) return 1;
      if (target->xrsl.Collect (*target, "opsys")) return 1;

      if (debug > 1) {
	cout << "Parsed output XRSL:" << endl;
	target->xrsl.Print();
      }

      // dump xrsl

      if (dumpxrsl) {
	cout << target->xrsl.str() << endl;
	submitted = true;
	continue;
      }

      // submit the jobs

      char * shortjobid = NULL;

      int err = ui_uploader (target->cluster->GetContact().c_str(),
			     target->xrsl.str().c_str(), &shortjobid,
			     target->cluster->GetContact().c_str(),
			     RSL_ACTION_REQUEST, filenames, debug);

      if (err || !shortjobid) {
	cout << "Job submission failed to queue " << target->queue->GetName()
	     << " at " << target->cluster->GetLongName() << endl;

	if (shortjobid) {
	  cout << "Deleting partially submitted job" << endl;
	  vector <string> filenames;
	  int err = ui_uploader (target->cluster->GetContact().c_str(), NULL,
				 &shortjobid, NULL, RSL_ACTION_CLEAN,
				 filenames, debug);
	  free (shortjobid);
	}

	cout << "Trying remaining queues" << endl;
	targetlist.erase (target);
	continue;
      }

      if (target->Accept()) return 1;

      string jobid = target->cluster->GetContact() + '/' + shortjobid;

      free (shortjobid);

      string jobsfilename = GetEnv ("HOME");
      jobsfilename.append ("/.ngjobs");
      ofstream ngjobs (jobsfilename.c_str(), ios::app);
      ngjobs << jobid << '#' << jobname << endl;
      ngjobs.close();

      string histfilename = GetEnv ("HOME");
      histfilename.append ("/.nghistory");
      ofstream nghist (histfilename.c_str(), ios::app);
      nghist << TimeStamp() << "  " << jobid << endl;
      nghist.close();

      if (!joblist.empty()) {
	ofstream joblist (joblist.c_str(), ios::app);
	joblist << jobid << endl;
	joblist.close();
      }

      cout << "Job submitted with jobid " << jobid << endl;

      submitted = true;
    }
  }

  return 0;
}
