#include "std.h"

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <getopt.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"
#include "Giis.h"

#ifdef __CYGWIN__
#define getopt_long_only getopt_long
#endif


void ngsyncusage (char * progname) {

  cout << "Usage: " << progname << " [options]" << endl;
  cout << endl;
  cout << "Options:" << endl;
  cout << "  -g, -giisurl   url           url to the central giis" << endl;
  cout << "  -t, -timeout   time          timeout for MDS queries in seconds (default " << DEFAULT_TIMEOUT << ")" << endl;
  cout << "  -d, -debug     debuglevel    0 = none, 1 = some, 2 = more, 3 = a lot" << endl;
  cout << "  -f, -force                   don't ask for verification" << endl;
  cout << "  -x, -anonymous               use anonymous bind for MDS queries (default)" << endl;
  cout << "  -X, -gsi                     use gsi-gssapi bind for MDS queries" << endl;
  cout << "  -v, -version                 print version information" << endl;
  cout << "  -h, -help                    print this help" << endl;
  cout << endl;
  cout << "The url to the giis has the format ldap://hostname[:port]/basedn" << endl;
  cout << "If the port is omitted the default 2135 is used" << endl;
  cout << "If the -g or -giisurl option is not given a list of urls is read from the" << endl;
  cout << "first existing file in the following list:" << endl;
  cout << "  1. $(HOME)/.nggiislist" << endl;
  cout << "  2. $(NORDUGRID_LOCATION)/etc/giislist" << endl;
  cout << "        or /opt/nordugrid/etc/giislist if $(NORDUGRID_LOCATION) is undefined" << endl;
  cout << "  3. /etc/giislist" << endl;
  cout << "Default values for the MDS timeout or the debuglevel can be set by defining" << endl;
  cout << "the environment variables NGTIMEOUT or NGDEBUG, or if these are undefined by" << endl;
  cout << "defining NGTIMEOUT or NGDEBUG in $(HOME)/.ngrc" << endl;
}


int ngsync (int argc, char ** argv) {

  // options

  vector <Giis> giislist;
  int timeout = UNDEFINED;
  int debug = UNDEFINED;
  bool force = false;
  bool anonymous = true;

  // parsing options

  char optstring[] = "g:t:d:fxXvh";
  struct option longoptions[] = {
    {"giisurl", 1, 0, 'g'},
    {"timeout", 1, 0, 't'},
    {"debug", 1, 0, 'd'},
    {"force", 0, 0, 'f'},
    {"anonymous", 0, 0, 'x'},
    {"gsi", 0, 0, 'X'},
    {"version", 0, 0, 'v'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  int opt;

  while (opt != -1) {

    opt = getopt_long_only (argc, argv, optstring, longoptions, NULL);

    switch (opt) {
    case -1:
      break;
    case 'g':
      {
	Giis giis (optarg);
	if (!giis) {
	  cerr << "Syntax error in giisurl argument: " << optarg << endl;
	  return 1;
	}
	else
	  giislist.push_back (giis);
      }
      break;
    case 't':
      timeout = atoi (optarg);
      break;
    case 'd':
      debug = atoi (optarg);
      break;
    case 'f':
      force = true;
      break;
    case 'x':
      anonymous = true;
      break;
    case 'X':
      anonymous = false;
      break;
    case 'v':
      cout << argv[0] << ": version " << VERSION << endl;
      return 0;
    case 'h':
      ngsyncusage (argv[0]);
      return 0;
    default:
      cerr << "Use \"" << argv[0] << " -help\" for help" << endl;
      return 1;
    }
  }

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) return 1;

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft() / 60) << endl;
  }

  if (!force) {
    cout << "Synchronizing the local list of active jobs with the information in the MDS" << endl;
    cout << "can result in some inconsistencies.  E.g. can very recently submitted jobs" << endl;
    cout << "not yet be present in the MDS information, and jobs very recently scheduled" << endl;
    cout << "for deletion might still be present in the MDS information." << endl;
    cout << "Are you sure you want to synchronize your local job list? [y/n] ";
    string response;
    cin >> response;
    if (response != "y" && response != "Y") {
      cout << "Cancelling synchronization request" << endl;
      return 0;
    }
  }

  // find available clusters

  if (giislist.empty()) if (GetGiises (giislist)) return 1;

  ClusterList cl = FindClusters (giislist, user.GetSN(), anonymous,
				 timeout, debug);
  if (cl.empty()) {
    cerr << argv[0] << ": could not retrieve cluster list from giis" << endl;
    return 1;
  }

  FindClusterInfo (cl, user.GetSN(), anonymous, timeout, debug);

  // finding the information

  string filename = GetEnv ("HOME");
  filename.append ("/.ngjobs");
  ofstream ngjobs (filename.c_str());

  if (debug) cout << "Found the following jobs:" << endl;

  for (ClusterList::iterator cli = cl.begin(); cli != cl.end(); cli++) {
    for (QueueList::iterator qli = cli->queues.begin();
	 qli != cli->queues.end(); qli++) {
      for (JobList::iterator jli = qli->jobs.begin();
	   jli != qli->jobs.end(); jli++) {
	ngjobs << jli->GetId() << '#' << jli-> GetName() << endl;
	if (debug) cout << jli->GetId() << endl;
      }
    }
  }

  ngjobs.close();

  return 0;
}
